# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for corpus_pruning_task."""
# pylint: disable=unused-argument
# pylint: disable=protected-access

import datetime
import os
import shutil
import tempfile
import unittest

import mock
import six

from clusterfuzz._internal.bot.fuzzers import options
from clusterfuzz._internal.bot.fuzzers.libFuzzer import \
    engine as libFuzzer_engine
from clusterfuzz._internal.bot.tasks import commands
from clusterfuzz._internal.bot.tasks import corpus_pruning_task
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.fuzzing import corpus_manager
from clusterfuzz._internal.google_cloud_utils import gsutil
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from clusterfuzz._internal.tests.test_libs import untrusted_runner_helpers

TEST_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), 'corpus_pruning_task_data')

TEST_GLOBAL_BUCKET = 'clusterfuzz-test-global-bundle'
TEST_SHARED_BUCKET = 'clusterfuzz-test-shared-corpus'
TEST2_BACKUP_BUCKET = 'clusterfuzz-test2-backup-bucket'


class BaseTest(object):
  """Base corpus pruning tests."""

  def setUp(self):
    """Setup."""
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.engine_common.unpack_seed_corpus_if_needed',
        'clusterfuzz._internal.bot.tasks.corpus_pruning_task.'
        'choose_cross_pollination_strategy',
        'clusterfuzz._internal.bot.tasks.task_creation.create_tasks',
        'clusterfuzz._internal.bot.tasks.setup.update_fuzzer_and_data_bundles',
        'clusterfuzz._internal.fuzzing.corpus_manager.backup_corpus',
        'clusterfuzz._internal.fuzzing.corpus_manager.GcsCorpus.rsync_to_disk',
        'clusterfuzz._internal.fuzzing.corpus_manager.FuzzTargetCorpus.rsync_from_disk',
        'clusterfuzz._internal.google_cloud_utils.blobs.write_blob',
        'clusterfuzz._internal.google_cloud_utils.storage.write_data',
        'clusterfuzz.fuzz.engine.get',
    ])
    self.mock.get.return_value = libFuzzer_engine.LibFuzzerEngine()
    self.mock.rsync_to_disk.side_effect = self._mock_rsync_to_disk
    self.mock.rsync_from_disk.side_effect = self._mock_rsync_from_disk
    self.mock.update_fuzzer_and_data_bundles.return_value = True
    self.mock.write_blob.return_value = 'key'
    self.mock.backup_corpus.return_value = 'backup_link'
    self.mock.choose_cross_pollination_strategy.return_value = ('random', None)

    def mocked_unpack_seed_corpus_if_needed(*args, **kwargs):
      """Mock's assert called methods are not powerful enough to ensure that
      unpack_seed_corpus_if_needed was called once with force_unpack=True.
      Instead, just assert that it was called once and during the call assert
      that it was called correctly.
      """
      self.assertTrue(kwargs.get('force_unpack', False))

    self.mock.unpack_seed_corpus_if_needed.side_effect = (
        mocked_unpack_seed_corpus_if_needed)

    data_types.FuzzTarget(
        engine='libFuzzer', binary='test_fuzzer', project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_test_fuzzer',
        engine='libFuzzer',
        job='libfuzzer_asan_job').put()

    self.fuzz_inputs_disk = tempfile.mkdtemp()
    self.bot_tmpdir = tempfile.mkdtemp()
    self.build_dir = os.path.join(TEST_DIR, 'build')
    self.corpus_bucket = tempfile.mkdtemp()
    self.corpus_dir = os.path.join(self.corpus_bucket, 'corpus')
    self.quarantine_dir = os.path.join(self.corpus_bucket, 'quarantine')
    self.shared_corpus_dir = os.path.join(self.corpus_bucket, 'shared')

    shutil.copytree(os.path.join(TEST_DIR, 'corpus'), self.corpus_dir)
    shutil.copytree(os.path.join(TEST_DIR, 'quarantine'), self.quarantine_dir)
    shutil.copytree(os.path.join(TEST_DIR, 'shared'), self.shared_corpus_dir)

    os.environ['BOT_TMPDIR'] = self.bot_tmpdir
    os.environ['FUZZ_INPUTS'] = self.fuzz_inputs_disk
    os.environ['FUZZ_INPUTS_DISK'] = self.fuzz_inputs_disk
    os.environ['CORPUS_BUCKET'] = 'bucket'
    os.environ['QUARANTINE_BUCKET'] = 'bucket-quarantine'
    os.environ['SHARED_CORPUS_BUCKET'] = 'bucket-shared'
    os.environ['JOB_NAME'] = 'libfuzzer_asan_job'
    os.environ['FAIL_RETRIES'] = '1'
    os.environ['APP_REVISION'] = '1337'

  def tearDown(self):
    shutil.rmtree(self.fuzz_inputs_disk, ignore_errors=True)
    shutil.rmtree(self.bot_tmpdir, ignore_errors=True)
    shutil.rmtree(self.corpus_bucket, ignore_errors=True)

  def _mock_setup_build(self, revision=None):
    os.environ['BUILD_DIR'] = self.build_dir
    return True

  def _mock_rsync_to_disk(self, _, sync_dir, timeout=None, delete=None):
    """Mock rsync_to_disk."""
    if 'quarantine' in sync_dir:
      corpus_dir = self.quarantine_dir
    elif 'shared' in sync_dir:
      corpus_dir = self.shared_corpus_dir
    else:
      corpus_dir = self.corpus_dir

    if os.path.exists(sync_dir):
      shutil.rmtree(sync_dir, ignore_errors=True)

    shutil.copytree(corpus_dir, sync_dir)
    return True

  def _mock_rsync_from_disk(self, _, sync_dir, timeout=None, delete=None):
    """Mock rsync_from_disk."""
    if 'quarantine' in sync_dir:
      corpus_dir = self.quarantine_dir
    else:
      corpus_dir = self.corpus_dir

    if os.path.exists(corpus_dir):
      shutil.rmtree(corpus_dir, ignore_errors=True)

    shutil.copytree(sync_dir, corpus_dir)
    return True


# TODO(unassigned): Support macOS.
@test_utils.supported_platforms('LINUX')
@test_utils.with_cloud_emulators('datastore')
class CorpusPruningTest(unittest.TestCase, BaseTest):
  """Corpus pruning tests."""

  def setUp(self):
    BaseTest.setUp(self)
    helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager.setup_build',
        'clusterfuzz._internal.base.utils.get_application_id',
    ])
    self.mock.setup_build.side_effect = self._mock_setup_build
    self.mock.get_application_id.return_value = 'project'

  def test_prune(self):
    """Basic pruning test."""
    corpus_pruning_task.execute_task('libFuzzer_test_fuzzer',
                                     'libfuzzer_asan_job')

    quarantined = os.listdir(self.quarantine_dir)
    self.assertEqual(1, len(quarantined))
    self.assertEqual(quarantined[0],
                     'crash-7acd6a2b3fe3c5ec97fa37e5a980c106367491fa')

    corpus = os.listdir(self.corpus_dir)
    self.assertEqual(4, len(corpus))
    six.assertCountEqual(self, [
        '39e0574a4abfd646565a3e436c548eeb1684fb57',
        '7d157d7c000ae27db146575c08ce30df893d3a64',
        '31836aeaab22dc49555a97edb4c753881432e01d',
        '6fa8c57336628a7d733f684dc9404fbd09020543',
    ], corpus)

    testcases = list(data_types.Testcase.query())
    self.assertEqual(1, len(testcases))
    self.assertEqual('Null-dereference WRITE', testcases[0].crash_type)
    self.assertEqual('Foo\ntest_fuzzer.cc\n', testcases[0].crash_state)
    self.assertEqual(1337, testcases[0].crash_revision)
    self.assertEqual('test_fuzzer',
                     testcases[0].get_metadata('fuzzer_binary_name'))
    self.assertEqual('label1,label2', testcases[0].get_metadata('issue_labels'))

    today = datetime.datetime.utcnow().date()
    # get_coverage_information on test_fuzzer rather than libFuzzer_test_fuzzer
    # since the libfuzzer_ prefix is removed when saving coverage info.
    coverage_info = data_handler.get_coverage_information('test_fuzzer', today)

    self.assertDictEqual(
        {
            'corpus_backup_location':
                u'backup_link',
            'corpus_location':
                u'gs://bucket/libFuzzer/test_fuzzer/',
            'corpus_size_bytes':
                8,
            'corpus_size_units':
                4,
            'date':
                today,
            # Coverage numbers are expected to be None as they come from fuzzer
            # coverage cron task (see src/go/server/cron/coverage.go).
            'edges_covered':
                None,
            'edges_total':
                None,
            'functions_covered':
                None,
            'functions_total':
                None,
            'fuzzer':
                u'test_fuzzer',
            'html_report_url':
                None,
            'quarantine_location':
                u'gs://bucket-quarantine/libFuzzer/test_fuzzer/',
            'quarantine_size_bytes':
                2,
            'quarantine_size_units':
                1,
        },
        coverage_info.to_dict())

    self.assertEqual(self.mock.unpack_seed_corpus_if_needed.call_count, 1)

  def test_get_libfuzzer_flags(self):
    """Test get_libfuzzer_flags logic."""
    fuzz_target = data_handler.get_fuzz_target('libFuzzer_test_fuzzer')
    context = corpus_pruning_task.Context(
        fuzz_target, [], corpus_pruning_task.Pollination.RANDOM, None)

    runner = corpus_pruning_task.Runner(self.build_dir, context)
    flags = runner.get_libfuzzer_flags()
    expected_default_flags = [
        '-timeout=5', '-rss_limit_mb=2560', '-max_len=5242880',
        '-detect_leaks=1', '-use_value_profile=1'
    ]
    six.assertCountEqual(self, flags, expected_default_flags)

    runner.fuzzer_options = options.FuzzerOptions(
        os.path.join(self.build_dir, 'test_get_libfuzzer_flags.options'))
    flags = runner.get_libfuzzer_flags()
    expected_custom_flags = [
        '-timeout=5', '-rss_limit_mb=2560', '-max_len=1337', '-detect_leaks=0',
        '-use_value_profile=1'
    ]
    six.assertCountEqual(self, flags, expected_custom_flags)


class CorpusPruningTestMinijail(CorpusPruningTest):
  """Tests for corpus pruning (minijail)."""

  def setUp(self):
    if environment.platform() != 'LINUX':
      self.skipTest('Minijail tests are only applicable for linux platform.')

    super(CorpusPruningTestMinijail, self).setUp()
    os.environ['USE_MINIJAIL'] = 'True'


@unittest.skipIf(
    not environment.get_value('FUCHSIA_TESTS'),
    'Temporarily disabling the Fuchsia test until build size reduced.')
@test_utils.with_cloud_emulators('datastore')
@test_utils.integration
class CorpusPruningTestFuchsia(unittest.TestCase, BaseTest):
  """Corpus pruning test for fuchsia."""

  def setUp(self):
    BaseTest.setUp(self)
    self.fuchsia_corpus_dir = os.path.join(self.corpus_bucket, 'fuchsia')
    shutil.copytree(os.path.join(TEST_DIR, 'fuchsia'), self.fuchsia_corpus_dir)
    self.temp_dir = tempfile.mkdtemp()
    builds_dir = os.path.join(self.temp_dir, 'builds')
    os.mkdir(builds_dir)
    urls_dir = os.path.join(self.temp_dir, 'urls')
    os.mkdir(urls_dir)

    environment.set_value('BUILDS_DIR', builds_dir)
    environment.set_value('BUILD_URLS_DIR', urls_dir)
    environment.set_value('QUEUE_OVERRIDE', 'FUCHSIA')
    environment.set_value('OS_OVERRIDE', 'FUCHSIA')

    env_string = ('RELEASE_BUILD_BUCKET_PATH = '
                  'gs://clusterfuchsia-builds-test/libfuzzer/'
                  'fuchsia-([0-9]+).zip')
    commands.update_environment_for_job(env_string)

    data_types.Job(
        name='libfuzzer_asan_fuchsia',
        platform='FUCHSIA',
        environment_string=env_string).put()
    data_types.FuzzTarget(
        binary='example-fuzzers/crash_fuzzer',
        engine='libFuzzer',
        project='fuchsia').put()

    environment.set_value('UNPACK_ALL_FUZZ_TARGETS_AND_FILES', True)
    helpers.patch(self, [
        'clusterfuzz._internal.system.shell.clear_temp_directory',
    ])

  def tearDown(self):
    shutil.rmtree(self.temp_dir, ignore_errors=True)

  def test_prune(self):
    """Basic pruning test."""
    self.corpus_dir = self.fuchsia_corpus_dir
    corpus_pruning_task.execute_task(
        'libFuzzer_fuchsia_example-fuzzers-crash_fuzzer',
        'libfuzzer_asan_fuchsia')
    corpus = os.listdir(self.corpus_dir)
    self.assertEqual(2, len(corpus))
    six.assertCountEqual(self, [
        '801c34269f74ed383fc97de33604b8a905adb635',
        '7cf184f4c67ad58283ecb19349720b0cae756829'
    ], corpus)
    quarantine = os.listdir(self.quarantine_dir)
    self.assertEqual(1, len(quarantine))
    six.assertCountEqual(
        self, ['crash-7a8dc3985d2a90fb6e62e94910fc11d31949c348'], quarantine)


class CorpusPruningTestUntrusted(
    untrusted_runner_helpers.UntrustedRunnerIntegrationTest):
  """Tests for corpus pruning (untrusted)."""

  def setUp(self):
    """Set up."""
    super(CorpusPruningTestUntrusted, self).setUp()
    environment.set_value('JOB_NAME', 'libfuzzer_asan_job')

    helpers.patch(self, [
        'clusterfuzz._internal.bot.tasks.setup.get_fuzzer_directory',
        'clusterfuzz._internal.base.tasks.add_task',
        'clusterfuzz._internal.bot.tasks.corpus_pruning_task.'
        '_record_cross_pollination_stats',
        'clusterfuzz.fuzz.engine.get',
    ])

    self.mock.get.return_value = libFuzzer_engine.LibFuzzerEngine()
    self.mock.get_fuzzer_directory.return_value = os.path.join(
        environment.get_value('ROOT_DIR'), 'src', 'clusterfuzz', '_internal',
        'bot', 'fuzzers', 'libFuzzer')
    self.corpus_bucket = os.environ['CORPUS_BUCKET']
    self.quarantine_bucket = os.environ['QUARANTINE_BUCKET']
    self.backup_bucket = os.environ['BACKUP_BUCKET']

    job = data_types.Job(
        name='libfuzzer_asan_job',
        environment_string=('APP_NAME = test_fuzzer\n'
                            'CORPUS_BUCKET = {corpus_bucket}\n'
                            'QUARANTINE_BUCKET = {quarantine_bucket}\n'
                            'BACKUP_BUCKET={backup_bucket}\n'
                            'RELEASE_BUILD_BUCKET_PATH = '
                            'gs://clusterfuzz-test-data/test_libfuzzer_builds/'
                            'test-libfuzzer-build-([0-9]+).zip\n'
                            'REVISION_VARS_URL = gs://clusterfuzz-test-data/'
                            'test_libfuzzer_builds/'
                            'test-libfuzzer-build-%s.srcmap.json\n'.format(
                                corpus_bucket=self.corpus_bucket,
                                quarantine_bucket=self.quarantine_bucket,
                                backup_bucket=self.backup_bucket)))
    job.put()

    job = data_types.Job(
        name='libfuzzer_asan_job2',
        environment_string=('APP_NAME = test2_fuzzer\n'
                            'BACKUP_BUCKET = {backup_bucket}\n'
                            'CORPUS_FUZZER_NAME_OVERRIDE = libfuzzer\n'.format(
                                backup_bucket=self.backup_bucket)))
    job.put()

    os.environ['PROJECT_NAME'] = 'oss-fuzz'
    data_types.FuzzTarget(
        engine='libFuzzer', project='test', binary='test_fuzzer').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_test_fuzzer',
        engine='libFuzzer',
        job='libfuzzer_asan_job',
        last_run=datetime.datetime.now()).put()

    data_types.FuzzTarget(
        engine='libFuzzer', project='test2', binary='fuzzer').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_test2_fuzzer',
        engine='libFuzzer',
        job='libfuzzer_asan_job2',
        last_run=datetime.datetime.now()).put()

    environment.set_value('USE_MINIJAIL', True)
    environment.set_value('SHARED_CORPUS_BUCKET', TEST_SHARED_BUCKET)

    # Set up remote corpora.
    self.corpus = corpus_manager.FuzzTargetCorpus('libFuzzer', 'test_fuzzer')
    self.corpus.rsync_from_disk(os.path.join(TEST_DIR, 'corpus'), delete=True)

    self.quarantine_corpus = corpus_manager.FuzzTargetCorpus(
        'libFuzzer', 'test_fuzzer', quarantine=True)
    self.quarantine_corpus.rsync_from_disk(
        os.path.join(TEST_DIR, 'quarantine'), delete=True)

    self.mock.get_data_bundle_bucket_name.return_value = TEST_GLOBAL_BUCKET
    data_types.DataBundle(
        name='bundle', is_local=True, sync_to_worker=True).put()

    data_types.Fuzzer(
        revision=1,
        file_size='builtin',
        source='builtin',
        name='libFuzzer',
        max_testcases=4,
        builtin=True,
        data_bundle_name='bundle').put()

    self.temp_dir = tempfile.mkdtemp()

    # Copy corpus backup in the older date format.
    corpus_backup_date = (
        datetime.datetime.utcnow().date() -
        datetime.timedelta(days=data_types.CORPUS_BACKUP_PUBLIC_LOOKBACK_DAYS))
    corpus_backup_dir = ('gs://{bucket}/corpus/libfuzzer/test2_fuzzer/')
    gsutil.GSUtilRunner().run_gsutil([
        'cp',
        (corpus_backup_dir + 'backup.zip').format(bucket=TEST2_BACKUP_BUCKET),
        (corpus_backup_dir +
         '%s.zip' % corpus_backup_date).format(bucket=self.backup_bucket)
    ])

  def tearDown(self):
    super(CorpusPruningTestUntrusted, self).tearDown()
    shutil.rmtree(self.temp_dir, ignore_errors=True)

  def test_prune(self):
    """Test pruning."""
    self._setup_env(job_type='libfuzzer_asan_job')
    self.mock._record_cross_pollination_stats.side_effect = (
        self.get_mock_record_compare(
            project_qualified_name='test_fuzzer',
            method='random',
            sources='test2_fuzzer',
            tags='',
            initial_corpus_size=5,
            corpus_size=3,
            initial_edge_coverage=0,
            edge_coverage=0,
            initial_feature_coverage=0,
            feature_coverage=0))

    corpus_pruning_task.execute_task('libFuzzer_test_fuzzer',
                                     'libfuzzer_asan_job')

    corpus_dir = os.path.join(self.temp_dir, 'corpus')
    os.mkdir(corpus_dir)
    self.corpus.rsync_to_disk(corpus_dir)

    six.assertCountEqual(self, [
        '39e0574a4abfd646565a3e436c548eeb1684fb57',
        '7d157d7c000ae27db146575c08ce30df893d3a64',
        '31836aeaab22dc49555a97edb4c753881432e01d',
        '6fa8c57336628a7d733f684dc9404fbd09020543',
    ], os.listdir(corpus_dir))

    quarantine_dir = os.path.join(self.temp_dir, 'quarantine')
    os.mkdir(quarantine_dir)
    self.quarantine_corpus.rsync_to_disk(quarantine_dir)

    six.assertCountEqual(self,
                         ['crash-7acd6a2b3fe3c5ec97fa37e5a980c106367491fa'],
                         os.listdir(quarantine_dir))

    testcases = list(data_types.Testcase.query())
    self.assertEqual(1, len(testcases))
    self.assertEqual('Null-dereference WRITE', testcases[0].crash_type)
    self.assertEqual('Foo\ntest_fuzzer.cc\n', testcases[0].crash_state)
    self.assertEqual(1337, testcases[0].crash_revision)
    self.assertEqual('test_fuzzer',
                     testcases[0].get_metadata('fuzzer_binary_name'))

    self.mock.add_task.assert_has_calls([
        mock.call('minimize', testcases[0].key.id(), u'libfuzzer_asan_job'),
    ])

    today = datetime.datetime.utcnow().date()
    coverage_info = data_handler.get_coverage_information('test_fuzzer', today)
    coverage_info_without_backup = coverage_info.to_dict()
    del coverage_info_without_backup['corpus_backup_location']

    self.assertDictEqual(
        {
            'corpus_location':
                u'gs://{}/libFuzzer/test_fuzzer/'.format(self.corpus_bucket),
            'corpus_size_bytes':
                8,
            'corpus_size_units':
                4,
            'date':
                today,
            # Coverage numbers are expected to be None as they come from fuzzer
            # coverage cron task (see src/go/server/cron/coverage.go).
            'edges_covered':
                None,
            'edges_total':
                None,
            'functions_covered':
                None,
            'functions_total':
                None,
            'fuzzer':
                u'test_fuzzer',
            'html_report_url':
                None,
            'quarantine_location':
                u'gs://{}/libFuzzer/test_fuzzer/'.format(self.quarantine_bucket
                                                        ),
            'quarantine_size_bytes':
                2,
            'quarantine_size_units':
                1,
        },
        coverage_info_without_backup)

    self.assertEqual(
        coverage_info.corpus_backup_location,
        'gs://{}/corpus/libFuzzer/test_fuzzer/'.format(
            self.backup_bucket) + '%s.zip' % today)

  def get_mock_record_compare(self, project_qualified_name, method, sources,
                              tags, initial_corpus_size, corpus_size,
                              initial_edge_coverage, edge_coverage,
                              initial_feature_coverage, feature_coverage):
    """Given all of the expected stats, returns a function
    that will compare them to an instance of CorpusPruningStats."""

    def compare(stats):
      """Mock record_cross_pollination_stats. Make sure function was called
      with the correct arguments."""
      self.assertEqual(project_qualified_name, stats.project_qualified_name)
      self.assertEqual(method, stats.method)
      self.assertEqual(tags, stats.tags)
      self.assertEqual(sources, stats.sources)
      self.assertEqual(initial_corpus_size, stats.initial_corpus_size)
      self.assertEqual(corpus_size, stats.corpus_size)
      self.assertEqual(initial_edge_coverage, stats.initial_edge_coverage)
      self.assertEqual(edge_coverage, stats.edge_coverage)
      self.assertEqual(initial_feature_coverage, stats.initial_feature_coverage)
      self.assertEqual(stats.feature_coverage, feature_coverage)

    return compare


@test_utils.with_cloud_emulators('datastore')
class CrossPollinationTest(unittest.TestCase):
  """Tests for cross pollination."""

  def test_select_targets_with_tagged_cross_pollination(self):
    """Test that selecting targets with a given tag returns the right target."""
    data_types.CorpusTag(
        tag='test_tag',
        fully_qualified_fuzz_target_name='libFuzzer_test_fuzzer').put()

    data_types.CorpusTag(
        tag='test_tag',
        fully_qualified_fuzz_target_name=
        'libFuzzer_cross_pollination_test_fuzzer').put()

    similar_target = data_types.FuzzTarget(
        engine='libFuzzer',
        binary='cross_pollination_test_fuzzer',
        project='test-project')
    similar_target.put()

    similar_job = data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_cross_pollination_test_fuzzer',
        engine='libFuzzer',
        job='libfuzzer_asan_job')
    similar_job.put()

    selected = corpus_pruning_task._select_targets_and_jobs_for_pollination(
        'libFuzzer', 'libFuzzer_test_fuzzer', 'tagged', 'test_tag')

    self.assertEqual([(similar_target, similar_job)], selected)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for untrusted."""

import os
import unittest

from clusterfuzz._internal.base.untrusted import untrusted_noop
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers


@untrusted_noop()
def test_function():
  return 42


@untrusted_noop(43)
def test_function2():
  return 42


class UntrustedNoopTest(unittest.TestCase):
  """Tests for untrusted_noop."""

  def setUp(self):
    test_helpers.patch_environ(self)

  def test_trusted(self):
    """Test calling function in trusted environment."""
    self.assertEqual(42, test_function())

  def test_untrusted(self):
    """Test calling function in untrusted environment."""
    os.environ['UNTRUSTED_WORKER'] = 'True'
    self.assertIsNone(test_function())
    self.assertEqual(43, test_function2())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Regression task.
   Find commit ranges where regressions were introduced."""

import random
import time

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.tasks import setup
from clusterfuzz._internal.bot.tasks import task_creation
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.build_management import revisions
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import big_query
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

# Number of revisions before the maximum to test before doing a bisect. This
# is also used as a cap for revisions to test near the minimum if the minimum
# happens to be a bad build.
EXTREME_REVISIONS_TO_TEST = 3

# Number of earlier revisions to check when validating ranges.
REVISIONS_TO_TEST_FOR_VALIDATION = 2

# Maximum revisions to look back when validating.
EARLIER_REVISIONS_TO_CONSIDER_FOR_VALIDATION = 10


def write_to_big_query(testcase, regression_range_start, regression_range_end):
  """Write the regression range to BigQuery."""
  big_query.write_range(
      table_id='regressions',
      testcase=testcase,
      range_name='regression',
      start=regression_range_start,
      end=regression_range_end)


def _save_current_regression_range_indices(testcase_id, regression_range_start,
                                           regression_range_end):
  """Save current regression range indices in case we die in middle of task."""
  testcase = data_handler.get_testcase_by_id(testcase_id)
  testcase.set_metadata(
      'last_regression_min', regression_range_start, update_testcase=False)
  testcase.set_metadata(
      'last_regression_max', regression_range_end, update_testcase=False)
  testcase.put()


def save_regression_range(testcase_id, regression_range_start,
                          regression_range_end):
  """Saves the regression range and creates blame and impact task if needed."""
  testcase = data_handler.get_testcase_by_id(testcase_id)
  testcase.regression = '%d:%d' % (regression_range_start, regression_range_end)
  data_handler.update_testcase_comment(
      testcase, data_types.TaskState.FINISHED,
      'regressed in range %s' % testcase.regression)

  write_to_big_query(testcase, regression_range_start, regression_range_end)

  # Force impacts update after regression range is updated. In several cases,
  # we might not have a production build to test with, so regression range is
  # used to decide impacts.
  task_creation.create_impact_task_if_needed(testcase)

  # Get blame information using the regression range result.
  task_creation.create_blame_task_if_needed(testcase)


def _testcase_reproduces_in_revision(testcase,
                                     testcase_file_path,
                                     job_type,
                                     revision,
                                     should_log=True,
                                     min_revision=None,
                                     max_revision=None):
  """Test to see if a test case reproduces in the specified revision."""
  if should_log:
    log_message = 'Testing r%d' % revision
    if min_revision is not None and max_revision is not None:
      log_message += ' (current range %d:%d)' % (min_revision, max_revision)

    testcase = data_handler.get_testcase_by_id(testcase.key.id())
    data_handler.update_testcase_comment(testcase, data_types.TaskState.WIP,
                                         log_message)

  build_manager.setup_build(revision)
  if not build_manager.check_app_path():
    raise errors.BuildSetupError(revision, job_type)

  if testcase_manager.check_for_bad_build(job_type, revision):
    log_message = 'Bad build at r%d. Skipping' % revision
    testcase = data_handler.get_testcase_by_id(testcase.key.id())
    data_handler.update_testcase_comment(testcase, data_types.TaskState.WIP,
                                         log_message)
    raise errors.BadBuildError(revision, job_type)

  test_timeout = environment.get_value('TEST_TIMEOUT', 10)
  result = testcase_manager.test_for_crash_with_retries(
      testcase, testcase_file_path, test_timeout, http_flag=testcase.http_flag)
  return result.is_crash()


def found_regression_near_extreme_revisions(testcase, testcase_file_path,
                                            job_type, revision_list, min_index,
                                            max_index):
  """Test to see if we regressed near either the min or max revision."""
  # Test a few of the most recent revisions.
  last_known_crashing_revision = revision_list[max_index]
  for offset in range(1, EXTREME_REVISIONS_TO_TEST + 1):
    current_index = max_index - offset
    if current_index < min_index:
      break

    # If we don't crash in a recent revision, we regressed in one of the
    # commits between the current revision and the one at the next index.
    try:
      is_crash = _testcase_reproduces_in_revision(
          testcase, testcase_file_path, job_type, revision_list[current_index])
    except errors.BadBuildError:
      # Skip this revision.
      continue

    if not is_crash:
      save_regression_range(testcase.key.id(), revision_list[current_index],
                            last_known_crashing_revision)
      return True

    last_known_crashing_revision = revision_list[current_index]

  # Test to see if we crash in the oldest revision we can run. This is a pre-
  # condition for our binary search. If we do crash in that revision, it
  # implies that we regressed between the first commit and our first revision,
  # which we represent as 0:|min_revision|.
  for _ in range(EXTREME_REVISIONS_TO_TEST):
    min_revision = revision_list[min_index]

    try:
      crashes_in_min_revision = _testcase_reproduces_in_revision(
          testcase,
          testcase_file_path,
          job_type,
          min_revision,
          should_log=False)
    except errors.BadBuildError:
      # If we find a bad build, potentially try another.
      if min_index + 1 >= max_index:
        break

      min_index += 1
      continue

    if crashes_in_min_revision:
      save_regression_range(testcase.key.id(), 0, min_revision)
      return True

    return False

  # We should have returned above. If we get here, it means we tried too many
  # builds near the min revision, and they were all bad.
  raise errors.BadBuildError(revision_list[min_index], job_type)


def validate_regression_range(testcase, testcase_file_path, job_type,
                              revision_list, min_index):
  """Ensure that we found the correct min revision by testing earlier ones."""
  earlier_revisions = revision_list[
      min_index - EARLIER_REVISIONS_TO_CONSIDER_FOR_VALIDATION:min_index]
  revision_count = min(len(earlier_revisions), REVISIONS_TO_TEST_FOR_VALIDATION)

  revisions_to_test = random.sample(earlier_revisions, revision_count)
  for revision in revisions_to_test:
    try:
      if _testcase_reproduces_in_revision(testcase, testcase_file_path,
                                          job_type, revision):
        testcase = data_handler.get_testcase_by_id(testcase.key.id())
        testcase.regression = 'NA'
        error_message = (
            'Low confidence in regression range. Test case crashes in '
            'revision r%d but not later revision r%d' %
            (revision, revision_list[min_index]))
        data_handler.update_testcase_comment(
            testcase, data_types.TaskState.ERROR, error_message)
        return False
    except errors.BadBuildError:
      pass

  return True


def find_regression_range(testcase_id, job_type):
  """Attempt to find when the testcase regressed."""
  deadline = tasks.get_task_completion_deadline()
  testcase = data_handler.get_testcase_by_id(testcase_id)
  if not testcase:
    return

  if testcase.regression:
    logs.log_error(
        'Regression range is already set as %s, skip.' % testcase.regression)
    return

  # This task is not applicable for custom binaries.
  if build_manager.is_custom_binary():
    testcase.regression = 'NA'
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         'Not applicable for custom binaries')
    return

  data_handler.update_testcase_comment(testcase, data_types.TaskState.STARTED)

  # Setup testcase and its dependencies.
  file_list, _, testcase_file_path = setup.setup_testcase(testcase, job_type)
  if not file_list:
    testcase = data_handler.get_testcase_by_id(testcase_id)
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         'Failed to setup testcase')
    tasks.add_task('regression', testcase_id, job_type)
    return

  build_bucket_path = build_manager.get_primary_bucket_path()
  revision_list = build_manager.get_revisions_list(
      build_bucket_path, testcase=testcase)
  if not revision_list:
    data_handler.close_testcase_with_error(testcase_id,
                                           'Failed to fetch revision list')
    return

  # Don't burden NFS server with caching these random builds.
  environment.set_value('CACHE_STORE', False)

  # Pick up where left off in a previous run if necessary.
  min_revision = testcase.get_metadata('last_regression_min')
  max_revision = testcase.get_metadata('last_regression_max')
  first_run = not min_revision and not max_revision
  if not min_revision:
    min_revision = revisions.get_first_revision_in_list(revision_list)
  if not max_revision:
    max_revision = testcase.crash_revision

  min_index = revisions.find_min_revision_index(revision_list, min_revision)
  if min_index is None:
    raise errors.BuildNotFoundError(min_revision, job_type)
  max_index = revisions.find_max_revision_index(revision_list, max_revision)
  if max_index is None:
    raise errors.BuildNotFoundError(max_revision, job_type)

  # Make sure that the revision where we noticed the crash, still crashes at
  # that revision. Otherwise, our binary search algorithm won't work correctly.
  max_revision = revision_list[max_index]
  crashes_in_max_revision = _testcase_reproduces_in_revision(
      testcase, testcase_file_path, job_type, max_revision, should_log=False)
  if not crashes_in_max_revision:
    testcase = data_handler.get_testcase_by_id(testcase_id)
    error_message = ('Known crash revision %d did not crash' % max_revision)
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         error_message)
    task_creation.mark_unreproducible_if_flaky(testcase, True)
    return

  # If we've made it this far, the test case appears to be reproducible. Clear
  # metadata from previous runs had it been marked as potentially flaky.
  task_creation.mark_unreproducible_if_flaky(testcase, False)

  # On the first run, check to see if we regressed near either the min or max
  # revision.
  if first_run and found_regression_near_extreme_revisions(
      testcase, testcase_file_path, job_type, revision_list, min_index,
      max_index):
    return

  while time.time() < deadline:
    min_revision = revision_list[min_index]
    max_revision = revision_list[max_index]

    # If the min and max revisions are one apart (or the same, if we only have
    # one build), this is as much as we can narrow the range.
    if max_index - min_index <= 1:
      # Verify that the regression range seems correct, and save it if so.
      if not validate_regression_range(testcase, testcase_file_path, job_type,
                                       revision_list, min_index):
        return

      save_regression_range(testcase_id, min_revision, max_revision)
      return

    middle_index = (min_index + max_index) // 2
    middle_revision = revision_list[middle_index]
    try:
      is_crash = _testcase_reproduces_in_revision(
          testcase,
          testcase_file_path,
          job_type,
          middle_revision,
          min_revision=min_revision,
          max_revision=max_revision)
    except errors.BadBuildError:
      # Skip this revision.
      del revision_list[middle_index]
      max_index -= 1
      continue

    if is_crash:
      max_index = middle_index
    else:
      min_index = middle_index

    _save_current_regression_range_indices(
        testcase_id, revision_list[min_index], revision_list[max_index])

  # If we've broken out of the above loop, we timed out. We'll finish by
  # running another regression task and picking up from this point.
  testcase = data_handler.get_testcase_by_id(testcase_id)
  error_message = 'Timed out, current range r%d:r%d' % (
      revision_list[min_index], revision_list[max_index])
  data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                       error_message)
  tasks.add_task('regression', testcase_id, job_type)


def execute_task(testcase_id, job_type):
  """Run regression task and handle potential errors."""
  try:
    find_regression_range(testcase_id, job_type)
  except errors.BuildSetupError as error:
    # If we failed to setup a build, it is likely a bot error. We can retry
    # the task in this case.
    testcase = data_handler.get_testcase_by_id(testcase_id)
    error_message = 'Build setup failed r%d' % error.revision
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         error_message)
    build_fail_wait = environment.get_value('FAIL_WAIT')
    tasks.add_task(
        'regression', testcase_id, job_type, wait_time=build_fail_wait)
  except errors.BadBuildError:
    # Though bad builds when narrowing the range are recoverable, certain builds
    # being marked as bad may be unrecoverable. Recoverable ones should not
    # reach this point.
    testcase = data_handler.get_testcase_by_id(testcase_id)
    testcase.regression = 'NA'
    error_message = 'Unable to recover from bad build'
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         error_message)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common functionality for engine fuzzers (ie: libFuzzer or AFL)."""

import contextlib
import glob
import json
import os
import pipes
import random
import re
import shutil
import sys
import time

import six

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.fuzzers import options
from clusterfuzz._internal.bot.fuzzers import utils as fuzzer_utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.fuzzing import strategy
from clusterfuzz._internal.metrics import fuzzer_stats
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import archive
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import minijail
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.system import shell

# Maximum allowed size of a corpus file.
CORPUS_INPUT_SIZE_LIMIT = 5 * 1024 * 1024

# Number of testcases to use for the corpus subset strategy.
# See https://crbug.com/682311 for more information.
# Size 100 has a slightly higher chance as it seems to be the best one so far.
CORPUS_SUBSET_NUM_TESTCASES = [10, 20, 50, 75, 75, 100, 100, 100, 125, 125, 150]

# Suffix used for seed corpus archive generated with build. Does not include
# file extension.
SEED_CORPUS_ARCHIVE_SUFFIX = '_seed_corpus'

# Maximum number of files in the corpus for which we will unpack the seed
# corpus.
MAX_FILES_FOR_UNPACK = 5

# Extension for owners file containing list of people to be notified.
OWNERS_FILE_EXTENSION = '.owners'

# Extension for per-fuzz target labels to be added to issue tracker.
LABELS_FILE_EXTENSION = '.labels'

# Extension for per-fuzz target components to be added to issue tracker.
COMPONENTS_FILE_EXTENSION = '.components'

# Extension for additional metadata to be added to issue descriptions.
METADATA_FILE_EXTENSION = '.issue_metadata'

# Header format for logs.
LOG_HEADER_FORMAT = ('Command: {command}\n' + 'Time ran: {time}\n')

# Number of radamsa mutations.
RADAMSA_MUTATIONS = 2000

# Maximum number of seconds to run radamsa for.
RADAMSA_TIMEOUT = 3


class Generator(object):
  """Generators we can use."""
  NONE = 0
  RADAMSA = 1
  ML_RNN = 2


def select_generator(strategy_pool, fuzzer_path):
  """Pick a generator to generate new testcases before fuzzing or return
  Generator.NONE if no generator selected."""
  if environment.is_lib() or environment.platform() == 'FUCHSIA':
    # Unsupported.
    return Generator.NONE

  # We can't use radamsa binary on Windows. Disable ML for now until we know it
  # works on Win.
  # These generators don't produce testcases that LPM fuzzers can use.
  if (environment.platform() == 'WINDOWS' or is_lpm_fuzz_target(fuzzer_path)):
    return Generator.NONE
  if strategy_pool.do_strategy(strategy.CORPUS_MUTATION_ML_RNN_STRATEGY):
    return Generator.ML_RNN
  if strategy_pool.do_strategy(strategy.CORPUS_MUTATION_RADAMSA_STRATEGY):
    return Generator.RADAMSA

  return Generator.NONE


def generate_new_testcase_mutations(corpus_directory,
                                    new_testcase_mutations_directory,
                                    fuzzer_name, candidate_generator):
  """Generate new testcase mutations, using existing corpus directory or other
  methods.

  Returns true if mutations are successfully generated using radamsa or ml rnn.
  A false return signifies either no generator use or unsuccessful generation of
  testcase mutations."""
  generation_timeout = get_new_testcase_mutations_timeout()
  pre_mutations_filecount = shell.get_directory_file_count(
      new_testcase_mutations_directory)

  # Generate new testcase mutations using Radamsa.
  if candidate_generator == Generator.RADAMSA:
    generate_new_testcase_mutations_using_radamsa(
        corpus_directory, new_testcase_mutations_directory, generation_timeout)
  # Generate new testcase mutations using ML RNN model.
  elif candidate_generator == Generator.ML_RNN:
    generate_new_testcase_mutations_using_ml_rnn(
        corpus_directory, new_testcase_mutations_directory, fuzzer_name,
        generation_timeout)

  # If new mutations are successfully generated, return true.
  if shell.get_directory_file_count(
      new_testcase_mutations_directory) > pre_mutations_filecount:
    return True

  return False


# Filename length limit on ext4.
FILENAME_LENGTH_LIMIT = 255

RADAMSA_FILENAME_REGEX = re.compile(r'radamsa-\d+-(.*)', re.DOTALL)


def get_radamsa_output_filename(initial_filename, i):
  """Get the name of a file mutated by radamsa."""
  # Don't add the radamsa prefix to a file that already has it to avoid hitting
  # filename/path length limits.
  match = RADAMSA_FILENAME_REGEX.search(initial_filename)
  if match:
    base_filename = match.group(1)
  else:
    base_filename = initial_filename
  prefix = 'radamsa-%05d-' % (i + 1)
  # FIXME: AFL will still break if the filename is near 255 chars since it
  # naively appends. AFL needs to rename every file to a sensible length (not
  # just those created by radamsa).
  return prefix + base_filename[:FILENAME_LENGTH_LIMIT - len(prefix)]


def generate_new_testcase_mutations_using_radamsa(
    corpus_directory, new_testcase_mutations_directory, generation_timeout):
  """Generate new testcase mutations based on Radamsa."""
  radamsa_path = get_radamsa_path()
  if not radamsa_path:
    # Mutations using radamsa are not supported on current platform, bail out.
    return

  radamsa_runner = new_process.ProcessRunner(radamsa_path)
  files_list = shell.get_files_list(corpus_directory)
  filtered_files_list = [
      f for f in files_list if os.path.getsize(f) <= CORPUS_INPUT_SIZE_LIMIT
  ]
  if not filtered_files_list:
    # No mutations to do on an empty corpus or one with very large files.
    return

  old_corpus_size = shell.get_directory_file_count(
      new_testcase_mutations_directory)
  expected_completion_time = time.time() + generation_timeout

  for i in range(RADAMSA_MUTATIONS):
    original_file_path = random_choice(filtered_files_list)
    original_filename = os.path.basename(original_file_path)
    output_path = os.path.join(
        new_testcase_mutations_directory,
        get_radamsa_output_filename(original_filename, i))

    result = radamsa_runner.run_and_wait(
        ['-o', output_path, original_file_path], timeout=RADAMSA_TIMEOUT)

    if (os.path.exists(output_path) and
        os.path.getsize(output_path) > CORPUS_INPUT_SIZE_LIMIT):
      # Skip large files to avoid further mutations and impact fuzzing
      # efficiency.
      shell.remove_file(output_path)
    elif result.return_code or result.timed_out:
      logs.log_warn(
          'Radamsa failed to mutate or timed out.', output=result.output)

    # Check if we exceeded our timeout. If yes, do no more mutations and break.
    if time.time() > expected_completion_time:
      break

  new_corpus_size = shell.get_directory_file_count(
      new_testcase_mutations_directory)
  logs.log('Added %d tests using Radamsa mutations.' %
           (new_corpus_size - old_corpus_size))


def generate_new_testcase_mutations_using_ml_rnn(
    corpus_directory, new_testcase_mutations_directory, fuzzer_name,
    generation_timeout):
  """Generate new testcase mutations using ML RNN model."""
  # No return value for now. Will add later if this is necessary.
  # Defer import to prevent issues with tensorflow causing hangs with
  # multiprocessing.
  try:
    from clusterfuzz._internal.bot.fuzzers.ml.rnn import \
        generator as ml_rnn_generator
  except Exception:
    logs.log_warn('Failed to import ml_rnn generator, skipping.')
    return

  ml_rnn_generator.execute(corpus_directory, new_testcase_mutations_directory,
                           fuzzer_name, generation_timeout)


def get_radamsa_path():
  """Return path to radamsa binary for current platform."""
  bin_directory_path = os.path.join(
      os.path.dirname(os.path.realpath(__file__)), 'bin')
  platform = environment.platform()
  if platform == 'LINUX':
    return os.path.join(bin_directory_path, 'linux', 'radamsa')

  if platform == 'MAC':
    return os.path.join(bin_directory_path, 'mac', 'radamsa')

  return None


def get_new_testcase_mutations_timeout():
  """Get the timeout for new testcase mutations."""
  return get_overridable_timeout(10 * 60, 'MUTATIONS_TIMEOUT_OVERRIDE')


def current_timestamp():
  """Returns the current timestamp. Needed for mocking."""
  return time.time()


def get_strategy_probability(strategy_name, default):
  """Returns a strategy weight based on env variable |FUZZING_STRATEGIES|"""
  fuzzing_strategies = environment.get_value('FUZZING_STRATEGIES')
  if fuzzing_strategies is None or not isinstance(fuzzing_strategies, dict):
    return default

  if strategy_name not in fuzzing_strategies:
    return 0.0

  return fuzzing_strategies[strategy_name]


def decide_with_probability(probability):
  """Decide if we want to do something with the given probability."""
  return random.SystemRandom().random() < probability


def get_probability():
  """Obtain a random number."""
  return random.SystemRandom().random()


def get_testcase_run(stats, fuzzer_command):
  """Get testcase run for stats."""
  build_revision = fuzzer_utils.get_build_revision()
  job = environment.get_value('JOB_NAME')
  # fuzzer name is filled by fuzz_task.
  testcase_run = fuzzer_stats.TestcaseRun(None, job, build_revision,
                                          current_timestamp())

  testcase_run['command'] = fuzzer_command
  testcase_run.update(stats)
  return testcase_run


def dump_big_query_data(stats, testcase_file_path, fuzzer_command):
  """Dump BigQuery stats."""
  testcase_run = get_testcase_run(stats, fuzzer_command)
  fuzzer_stats.TestcaseRun.write_to_disk(testcase_run, testcase_file_path)


def find_fuzzer_path(build_directory, fuzzer_name, is_blackbox=False):
  """Find the fuzzer path with the given name."""
  # Blackbox fuzzers are special cases. They run from the fuzzers directory
  # rather than using a target from the build archive.
  if is_blackbox:
    fuzzer_directory = environment.get_value('FUZZERS_DIR')
    fuzzer_directory = os.path.join(fuzzer_directory, fuzzer_name)
    fuzzer = data_types.Fuzzer.query(
        data_types.Fuzzer.name == fuzzer_name).get()
    return os.path.join(fuzzer_directory, fuzzer.executable_path)

  if not build_directory:
    # Grey-box fuzzers might not have the build directory for a particular job
    # configuration when doing variant task testing (e.g. Android on-device
    # fuzz target might not exist on host). In this case, treat it similar to
    # target not found by returning None.
    logs.log_warn('No build directory found for fuzzer: %s' % fuzzer_name)
    return None

  if environment.platform() == 'FUCHSIA':
    # Fuchsia targets are not on disk.
    return fuzzer_name

  if environment.is_android_kernel():
    return os.path.join(build_directory, 'syzkaller', 'bin', 'syz-manager')

  # TODO(ochang): This is necessary for legacy testcases, which include the
  # project prefix in arguments. Remove this in the near future.
  project_name = environment.get_value('PROJECT_NAME')
  legacy_name_prefix = u''
  if project_name:
    legacy_name_prefix = project_name + u'_'

  fuzzer_filename = environment.get_executable_filename(fuzzer_name)
  for root, _, files in shell.walk(build_directory):
    for filename in files:
      if (legacy_name_prefix + filename == fuzzer_name or
          filename == fuzzer_filename):
        return os.path.join(root, filename)

  # This is an expected case when doing regression testing with old builds
  # that do not have that fuzz target. It can also happen when a host sends a
  # message to an untrusted worker that just restarted and lost information on
  # build directory.
  logs.log_warn('Fuzzer: %s not found in build_directory: %s.' %
                (fuzzer_name, build_directory))
  return None


def get_command_quoted(command):
  """Return shell quoted command string."""
  return ' '.join(pipes.quote(part) for part in command)


def get_overridable_timeout(default_timeout, override_env_var):
  """Returns a timeout given a |default_timeout| and the environment variable,
  |override_env_var|, that overrides it. Returns the overriden value if
  |override_env_var| is set, otherwise returns default_timeout. Throws an
  assertion error if the return value is negative."""
  timeout_override = environment.get_value(override_env_var)
  timeout = float(timeout_override or default_timeout)
  assert timeout >= 0, timeout
  return timeout


def get_hard_timeout(total_timeout=None):
  """Get the hard timeout for fuzzing."""
  if total_timeout is None:
    total_timeout = environment.get_value('FUZZ_TEST_TIMEOUT')

  return get_overridable_timeout(total_timeout, 'HARD_TIMEOUT_OVERRIDE')


def get_merge_timeout(default_merge_timeout):
  """Get the maximum amount of time that should be spent merging a corpus."""
  return get_overridable_timeout(default_merge_timeout,
                                 'MERGE_TIMEOUT_OVERRIDE')


def is_lpm_fuzz_target(fuzzer_path):
  """Returns True if |fuzzer_path| is a libprotobuf-mutator based fuzz
  target."""
  # TODO(metzman): Use this function to disable running LPM targets with AFL.
  with open(fuzzer_path, 'rb') as file_handle:
    return utils.search_bytes_in_file(b'TestOneProtoInput', file_handle)


def get_issue_owners(fuzz_target_path):
  """Return list of owner emails given a fuzz target path.

  Format of an owners file is described at:
  https://cs.chromium.org/chromium/src/third_party/depot_tools/owners.py
  """
  owners_file_path = fuzzer_utils.get_supporting_file(fuzz_target_path,
                                                      OWNERS_FILE_EXTENSION)

  if environment.is_trusted_host():
    owners_file_path = fuzzer_utils.get_file_from_untrusted_worker(
        owners_file_path)

  if not os.path.exists(owners_file_path):
    return []

  owners = []
  with open(owners_file_path, 'r') as owners_file_handle:
    owners_file_content = owners_file_handle.read()

    for line in owners_file_content.splitlines():
      stripped_line = line.strip()
      if not stripped_line:
        # Ignore empty lines.
        continue
      if stripped_line.startswith('#'):
        # Ignore comment lines.
        continue
      if stripped_line == '*':
        # Not of any use, we can't add everyone as owner with this.
        continue
      if (stripped_line.startswith('per-file') or
          stripped_line.startswith('file:')):
        # Don't have a source checkout, so ignore.
        continue
      if '@' not in stripped_line:
        # Bad email address.
        continue
      owners.append(stripped_line)

  return owners


def get_issue_metadata(fuzz_target_path, extension):
  """Get issue metadata."""
  metadata_file_path = fuzzer_utils.get_supporting_file(fuzz_target_path,
                                                        extension)

  if environment.is_trusted_host():
    metadata_file_path = fuzzer_utils.get_file_from_untrusted_worker(
        metadata_file_path)

  if not os.path.exists(metadata_file_path):
    return []

  with open(metadata_file_path) as handle:
    return utils.parse_delimited(
        handle, delimiter='\n', strip=True, remove_empty=True)


def get_issue_labels(fuzz_target_path):
  """Return list of issue labels given a fuzz target path."""
  return get_issue_metadata(fuzz_target_path, LABELS_FILE_EXTENSION)


def get_issue_components(fuzz_target_path):
  """Return list of issue components given a fuzz target path."""
  return get_issue_metadata(fuzz_target_path, COMPONENTS_FILE_EXTENSION)


def get_additional_issue_metadata(fuzz_target_path):
  """Return the additional metadata fields given a fuzz target path. The data
  will be a JSON-formatted dictionary."""
  metadata_file_path = fuzzer_utils.get_supporting_file(
      fuzz_target_path, METADATA_FILE_EXTENSION)

  if environment.is_trusted_host():
    metadata_file_path = fuzzer_utils.get_file_from_untrusted_worker(
        metadata_file_path)

  if not os.path.exists(metadata_file_path):
    return {}

  with open(metadata_file_path) as handle:
    try:
      return json.load(handle)
    except (ValueError, TypeError):
      logs.log_error('Invalid metadata file format.', path=metadata_file_path)
      return {}


def get_all_issue_metadata(fuzz_target_path):
  """Get issue related metadata for a target."""
  metadata = {}

  issue_labels = get_issue_labels(fuzz_target_path)
  if issue_labels:
    metadata['issue_labels'] = ','.join(issue_labels)

  issue_components = get_issue_components(fuzz_target_path)
  if issue_components:
    metadata['issue_components'] = ','.join(issue_components)

  issue_owners = get_issue_owners(fuzz_target_path)
  if issue_owners:
    metadata['issue_owners'] = ','.join(issue_owners)

  issue_metadata = get_additional_issue_metadata(fuzz_target_path)
  if issue_metadata:
    metadata['issue_metadata'] = issue_metadata

  return metadata


def get_all_issue_metadata_for_testcase(testcase):
  """Get issue related metadata given a testcase."""
  if environment.is_trusted_host():
    # Not applicable.
    return None

  fuzz_target = testcase.get_fuzz_target()
  if not fuzz_target:
    return None

  build_dir = environment.get_value('BUILD_DIR')
  is_blackbox = fuzz_target.engine == 'blackbox'
  target_path = find_fuzzer_path(
      build_dir, fuzz_target.binary, is_blackbox=is_blackbox)
  if not target_path:
    logs.log_error('Failed to find target path for ' + fuzz_target.binary)
    return None

  return get_all_issue_metadata(target_path)


def get_project_qualified_fuzzer_name(target_path):
  """Return project qualified fuzzer name for a given target path."""
  return data_types.fuzz_target_project_qualified_name(
      utils.current_project(), os.path.basename(target_path))


def format_fuzzing_strategies(fuzzing_strategies):
  """Format the strategies used for logging purposes."""
  if not fuzzing_strategies:
    return ''

  if isinstance(fuzzing_strategies, list):
    # Legacy format. TODO(ochang): Remove this once it's not used.
    value = ','.join(fuzzing_strategies)
  else:
    # New format.
    assert isinstance(fuzzing_strategies, dict)
    value = ','.join('{}:{}'.format(key, value)
                     for key, value in six.iteritems(fuzzing_strategies))

  return 'cf::fuzzing_strategies: ' + value


def random_choice(sequence):
  """Return a random element from the non-empty sequence."""
  return random.SystemRandom().choice(sequence)


def read_data_from_file(file_path):
  """Read data from file."""
  with open(file_path, 'rb') as file_handle:
    return file_handle.read()


def recreate_directory(directory_path):
  """Delete directory if exists, create empty directory. Throw an exception if
  either fails."""
  if not shell.remove_directory(directory_path, recreate=True):
    raise Exception('Failed to recreate directory: ' + directory_path)


def strip_minijail_command(command, fuzzer_path):
  """Remove minijail arguments from a fuzzer command.

  Args:
    command: The command.
    fuzzer_path: Absolute path to the fuzzer.

  Returns:
    The stripped command.
  """
  try:
    fuzzer_path_index = command.index(fuzzer_path)
    return command[fuzzer_path_index:]
  except ValueError:
    return command


def write_data_to_file(content, file_path):
  """Writes data to file."""
  with open(file_path, 'wb') as file_handle:
    file_handle.write(str(content).encode('utf-8'))


class MinijailEngineFuzzerRunner(minijail.MinijailProcessRunner):
  """Minijail runner for engine fuzzers."""

  @contextlib.contextmanager
  def _chroot_testcase(self, testcase_path):
    """Context manager for testcases.
    Args:
      testcase_path: Host path to testcase.
    Yields:
      Path to testcase within chroot.
    """
    testcase_directory, testcase_name = os.path.split(testcase_path)
    binding = self.chroot.get_binding(testcase_directory)
    if binding:
      # The host directory that contains this testcase is bound in the chroot.
      yield os.path.join(binding.dest_path, testcase_name)
      return
    # Copy the testcase into the chroot (temporarily).
    shutil.copy(testcase_path, self.chroot.directory)
    copied_testcase_path = os.path.join(self.chroot.directory, testcase_name)
    yield '/' + testcase_name
    # Cleanup
    os.remove(copied_testcase_path)


def signal_term_handler(sig, frame):  # pylint: disable=unused-argument
  try:
    print('SIGTERMed')
  except IOError:  # Pipe may already be closed and we may not be able to print.
    pass

  new_process.kill_process_tree(os.getpid())
  sys.exit(0)


def get_seed_corpus_path(fuzz_target_path):
  """Returns the path of the seed corpus if one exists. Otherwise returns None.
  Logs an error if multiple seed corpora exist for the same target."""
  archive_path_without_extension = fuzzer_utils.get_supporting_file(
      fuzz_target_path, SEED_CORPUS_ARCHIVE_SUFFIX)
  # Get all files that end with _seed_corpus.*
  possible_archive_paths = set(glob.glob(archive_path_without_extension + '.*'))
  # Now get a list of these that are valid seed corpus archives.
  archive_paths = possible_archive_paths.intersection(
      set(archive_path_without_extension + extension
          for extension in archive.ARCHIVE_FILE_EXTENSIONS))

  archive_paths = list(archive_paths)
  if not archive_paths:
    return None

  if len(archive_paths) > 1:
    logs.log_error('Multiple seed corpuses exist for fuzz target %s: %s.' %
                   (fuzz_target_path, ', '.join(archive_paths)))

  return archive_paths[0]


def process_sanitizer_options_overrides(fuzzer_path):
  """Applies sanitizer option overrides from .options file."""
  fuzzer_options = options.get_fuzz_target_options(fuzzer_path)
  if not fuzzer_options:
    return

  asan_options = environment.get_memory_tool_options('ASAN_OPTIONS', {})
  msan_options = environment.get_memory_tool_options('MSAN_OPTIONS', {})
  ubsan_options = environment.get_memory_tool_options('UBSAN_OPTIONS', {})
  hwasan_options = environment.get_memory_tool_options('HWASAN_OPTIONS', {})

  asan_overrides = fuzzer_options.get_asan_options()
  if asan_options and asan_overrides:
    asan_options.update(asan_overrides)
    environment.set_memory_tool_options('ASAN_OPTIONS', asan_options)

  msan_overrides = fuzzer_options.get_msan_options()
  if msan_options and msan_overrides:
    msan_options.update(msan_overrides)
    environment.set_memory_tool_options('MSAN_OPTIONS', msan_options)

  ubsan_overrides = fuzzer_options.get_ubsan_options()
  if ubsan_options and ubsan_overrides:
    ubsan_options.update(ubsan_overrides)
    environment.set_memory_tool_options('UBSAN_OPTIONS', ubsan_options)

  hwasan_overrides = fuzzer_options.get_hwasan_options()
  if hwasan_options and hwasan_overrides:
    hwasan_options.update(hwasan_overrides)
    environment.set_memory_tool_options('HWASAN_OPTIONS', hwasan_options)


def unpack_seed_corpus_if_needed(fuzz_target_path,
                                 corpus_directory,
                                 max_bytes=float('inf'),
                                 force_unpack=False,
                                 max_files_for_unpack=MAX_FILES_FOR_UNPACK):
  """If seed corpus available, unpack it into the corpus directory if needed,
  ie: if corpus exists and either |force_unpack| is True, or the number of files
  in corpus_directory is less than |max_files_for_unpack|. Uses
  |fuzz_target_path| to find the seed corpus. If max_bytes is specified, then
  seed corpus files larger than |max_bytes| will not be unpacked.
  """
  seed_corpus_archive_path = get_seed_corpus_path(fuzz_target_path)
  if not seed_corpus_archive_path:
    return

  num_corpus_files = len(shell.get_files_list(corpus_directory))
  if not force_unpack and num_corpus_files > max_files_for_unpack:
    return

  if force_unpack:
    logs.log('Forced unpack: %s.' % seed_corpus_archive_path)

  archive_iterator = archive.iterator(seed_corpus_archive_path)
  # Unpack seed corpus recursively into the root of the main corpus directory.
  idx = 0
  for seed_corpus_file in archive_iterator:
    # Ignore directories.
    if seed_corpus_file.name.endswith('/'):
      continue

    # Allow callers to opt-out of unpacking large files.
    if seed_corpus_file.size > max_bytes:
      continue

    output_filename = '%016d' % idx
    output_file_path = os.path.join(corpus_directory, output_filename)
    with open(output_file_path, 'wb') as file_handle:
      shutil.copyfileobj(seed_corpus_file.handle, file_handle)

    idx += 1

  logs.log('Unarchiving %d files from seed corpus %s.' %
           (idx, seed_corpus_archive_path))


def get_log_header(command, time_executed):
  """Get the log header."""
  return LOG_HEADER_FORMAT.format(
      command=get_command_quoted(command), time=time_executed)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler that uploads a testcase"""

import ast
import datetime
import io
import json
import os

from flask import request
from google.cloud import ndb

from clusterfuzz._internal.base import external_users
from clusterfuzz._internal.base import memoize
from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.crash_analysis.stack_parsing import stack_analyzer
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import blobs
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.system import archive
from clusterfuzz._internal.system import environment
from handlers import base_handler
from libs import access
from libs import form
from libs import gcs
from libs import handler
from libs import helpers
from libs.issue_management import issue_tracker_utils
from libs.query import datastore_query

MAX_RETRIES = 50
RUN_FILE_PATTERNS = ['run', 'fuzz-', 'index.', 'crash.']
PAGE_SIZE = 20
MORE_LIMIT = 100 - PAGE_SIZE
UPLOAD_URL = '/upload-testcase/upload-oauth'
MEMCACHE_TTL_IN_SECONDS = 60 * 60  # 1 hour.


def _is_uploader_allowed(email):
  """Return bool on whether user is allowed to upload to any job or fuzzer."""
  return external_users.is_upload_allowed_for_user(email)


def attach_testcases(rows):
  """Attach testcase to each crash."""
  testcases = {}
  for index, row in enumerate(rows):
    testcases[index] = query_testcase(row['testcaseId'])

  for index, row in enumerate(rows):
    testcase = (list(testcases[index]) or [None])[0]
    if testcase:
      testcase = {
          'crashType': testcase.crash_type,
          'crashStateLines': (testcase.crash_state or '').strip().splitlines(),
          'isSecurity': testcase.security_flag,
          'issueNumber': testcase.bug_information,
          'job': testcase.job_type,
          'fuzzerName': testcase.actual_fuzzer_name(),
          'projectName': testcase.project_name,
      }
    row['testcase'] = testcase


def get_result():
  """Get the result."""
  params = dict(request.iterparams())
  page = helpers.cast(request.get('page') or 1, int, "'page' is not an int.")

  query = datastore_query.Query(data_types.TestcaseUploadMetadata)
  query.order('timestamp', is_desc=True)

  if not access.has_access(need_privileged_access=True):
    query.filter('uploader_email', helpers.get_user_email())
    params['permission'] = {'uploaderEmail': helpers.get_user_email()}

  entities, total_pages, total_items, has_more = query.fetch_page(
      page=page, page_size=PAGE_SIZE, projection=None, more_limit=MORE_LIMIT)

  items = []
  for entity in entities:
    items.append({
        'timestamp': utils.utc_datetime_to_timestamp(entity.timestamp),
        'testcaseId': entity.testcase_id,
        'uploaderEmail': entity.uploader_email,
        'filename': entity.filename,
        'bundled': entity.bundled,
        'pathInArchive': entity.path_in_archive,
        'status': entity.status
    })

  attach_testcases(items)

  result = {
      'hasMore': has_more,
      'items': items,
      'page': page,
      'pageSize': PAGE_SIZE,
      'totalItems': total_items,
      'totalPages': total_pages,
  }
  return result, params


def _read_to_bytesio(gcs_path):
  """Return a bytesio representing a GCS object."""
  data = storage.read_data(gcs_path)
  if not data:
    raise helpers.EarlyExitException('Failed to read uploaded archive.', 500)

  return io.BytesIO(data)


def guess_input_file(uploaded_file, filename):
  """Guess the main test case file from an archive."""
  for file_pattern in RUN_FILE_PATTERNS:
    blob_reader = _read_to_bytesio(uploaded_file.gcs_path)
    file_path_input = archive.get_first_file_matching(file_pattern, blob_reader,
                                                      filename)
    if file_path_input:
      return file_path_input

  return None


def query_testcase(testcase_id):
  """Start a query for an associated testcase."""
  if not testcase_id:
    return []

  return data_types.Testcase.query(data_types.Testcase.key == ndb.Key(
      data_types.Testcase, testcase_id)).iter(
          limit=1,
          projection=[
              'crash_type',
              'crash_state',
              'security_flag',
              'bug_information',
              'job_type',
              'fuzzer_name',
              'overridden_fuzzer_name',
              'project_name',
          ])


def filter_target_names(targets, engine):
  """Filter target names for a fuzzer and remove parent fuzzer prefixes."""
  prefix = engine + '_'
  return [t[len(prefix):] for t in targets if t.startswith(prefix)]


def filter_blackbox_fuzzers(fuzzers):
  """Filter out fuzzers such that only blackbox fuzzers are included."""

  def is_greybox_fuzzer(name):
    # TODO(ochang): Generalize this check.
    return (name.startswith('libFuzzer') or name.startswith('afl') or
            name.startswith('honggfuzz'))

  return [f for f in fuzzers if not is_greybox_fuzzer(f)]


@memoize.wrap(memoize.Memcache(MEMCACHE_TTL_IN_SECONDS))
def find_fuzz_target(engine, target_name, job_name):
  """Return fuzz target values given the engine, target name (which may or may
  not be prefixed with project), and job."""
  project_name = data_handler.get_project_name(job_name)
  candidate_name = data_types.fuzz_target_fully_qualified_name(
      engine, project_name, target_name)

  target = data_handler.get_fuzz_target(candidate_name)
  if not target:
    raise helpers.EarlyExitException('Fuzz target does not exist.', 400)

  return target.fully_qualified_name(), target.binary


def _allow_unprivileged_metadata(testcase_metadata):
  """Returns whether or not the provided testcase metadata can be set by an
  unprivileged user."""
  if utils.is_oss_fuzz():
    # Labels in OSS-Fuzz are privileged and control things like disclosure
    # deadlines. Do not let these be editable.
    return False

  # Allow *only* issue labels to be set.
  return len(testcase_metadata) == 1 and 'issue_labels' in testcase_metadata


class Handler(base_handler.Handler):
  """Handler for the testcase uploads page."""

  @handler.get(handler.HTML)
  def get(self):
    """Handles get request."""
    email = helpers.get_user_email()
    if not email:
      raise helpers.AccessDeniedException()

    is_privileged_or_domain_user = access.has_access(
        need_privileged_access=False)
    if is_privileged_or_domain_user or _is_uploader_allowed(email):
      # Privileged, domain and upload users can see all job and fuzzer names.
      allowed_jobs = data_handler.get_all_job_type_names()
      allowed_fuzzers = data_handler.get_all_fuzzer_names_including_children(
          include_parents=True)
    else:
      # Check if this is an external user with access to certain fuzzers/jobs.
      allowed_jobs = external_users.allowed_jobs_for_user(email)
      allowed_fuzzers = external_users.allowed_fuzzers_for_user(
          email, include_from_jobs=True)

      if not allowed_fuzzers and not allowed_jobs:
        raise helpers.AccessDeniedException()

    has_issue_tracker = bool(data_handler.get_issue_tracker_name())

    result, params = get_result()
    return self.render(
        'upload.html', {
            'fieldValues': {
                'blackboxFuzzers':
                    filter_blackbox_fuzzers(allowed_fuzzers),
                'jobs':
                    allowed_jobs,
                'libfuzzerTargets':
                    filter_target_names(allowed_fuzzers, 'libFuzzer'),
                'aflTargets':
                    filter_target_names(allowed_fuzzers, 'afl'),
                'honggfuzzTargets':
                    filter_target_names(allowed_fuzzers, 'honggfuzz'),
                'isChromium':
                    utils.is_chromium(),
                'sandboxedJobs':
                    data_types.INTERNAL_SANDBOXED_JOB_TYPES,
                'csrfToken':
                    form.generate_csrf_token(),
                'isExternalUser':
                    not is_privileged_or_domain_user,
                'uploadInfo':
                    gcs.prepare_blob_upload()._asdict(),
                'hasIssueTracker':
                    has_issue_tracker,
            },
            'params': params,
            'result': result
        })


class PrepareUploadHandler(base_handler.Handler):
  """Handler that creates an upload URL."""

  @handler.check_user_access(need_privileged_access=False)
  def post(self):
    """Serves the url."""
    return self.render_json({'uploadInfo': gcs.prepare_blob_upload()._asdict()})


class UploadUrlHandlerOAuth(base_handler.Handler):
  """Handler that creates an upload URL (OAuth)."""

  @handler.oauth
  @handler.check_user_access(need_privileged_access=False)
  def post(self):
    """Serves the url."""
    return self.render_json({
        'uploadUrl': request.host_url + UPLOAD_URL,
    })


class JsonHandler(base_handler.Handler):
  """JSON handler for past testcase uploads."""

  @handler.post(handler.JSON, handler.JSON)
  def post(self):
    """Handles a post request."""
    if not helpers.get_user_email():
      raise helpers.AccessDeniedException()

    result, _ = get_result()
    return self.render_json(result)


class UploadHandlerCommon(object):
  """Handler that uploads the testcase file."""

  def get_upload(self):
    """Get the upload."""
    raise NotImplementedError

  def do_post(self):
    """Upload a testcase."""
    email = helpers.get_user_email()
    testcase_id = request.get('testcaseId')
    uploaded_file = self.get_upload()
    if testcase_id and not uploaded_file:
      testcase = helpers.get_testcase(testcase_id)
      if not access.can_user_access_testcase(testcase):
        raise helpers.AccessDeniedException()

      # Use minimized testcase for upload (if available).
      key = (
          testcase.minimized_keys if testcase.minimized_keys and
          testcase.minimized_keys != 'NA' else testcase.fuzzed_keys)

      uploaded_file = blobs.get_blob_info(key)

      # Extract filename part from blob.
      uploaded_file.filename = os.path.basename(
          uploaded_file.filename.replace('\\', os.sep))

    job_type = request.get('job')
    if not job_type:
      raise helpers.EarlyExitException('Missing job name.', 400)

    job = data_types.Job.query(data_types.Job.name == job_type).get()
    if not job:
      raise helpers.EarlyExitException('Invalid job name.', 400)

    fuzzer_name = request.get('fuzzer')
    job_type_lowercase = job_type.lower()
    if 'libfuzzer' in job_type_lowercase:
      fuzzer_name = 'libFuzzer'
    elif 'afl' in job_type_lowercase:
      fuzzer_name = 'afl'
    elif 'honggfuzz' in job_type_lowercase:
      fuzzer_name = 'honggfuzz'

    is_engine_job = fuzzer_name and environment.is_engine_fuzzer_job(job_type)
    target_name = request.get('target')
    if not is_engine_job and target_name:
      raise helpers.EarlyExitException(
          'Target name is not applicable to non-engine jobs (AFL, libFuzzer).',
          400)

    if is_engine_job and not target_name:
      raise helpers.EarlyExitException(
          'Missing target name for engine job (AFL, libFuzzer).', 400)

    if (target_name and
        not data_types.Fuzzer.VALID_NAME_REGEX.match(target_name)):
      raise helpers.EarlyExitException('Invalid target name.', 400)

    fully_qualified_fuzzer_name = ''
    if is_engine_job and target_name:
      if job.is_external():
        # External jobs don't run and set FuzzTarget entities as part of
        # fuzz_task. Set it here instead.
        fuzz_target = (
            data_handler.record_fuzz_target(fuzzer_name, target_name, job_type))
        fully_qualified_fuzzer_name = fuzz_target.fully_qualified_name()
        target_name = fuzz_target.binary
      else:
        fully_qualified_fuzzer_name, target_name = find_fuzz_target(
            fuzzer_name, target_name, job_type)

    if (not access.has_access(
        need_privileged_access=False,
        job_type=job_type,
        fuzzer_name=(fully_qualified_fuzzer_name or fuzzer_name)) and
        not _is_uploader_allowed(email)):
      raise helpers.AccessDeniedException()

    multiple_testcases = bool(request.get('multiple'))
    http_flag = bool(request.get('http'))
    high_end_job = bool(request.get('highEnd'))
    bug_information = request.get('issue')
    crash_revision = request.get('revision')
    timeout = request.get('timeout')
    retries = request.get('retries')
    bug_summary_update_flag = bool(request.get('updateIssue'))
    quiet_flag = bool(request.get('quiet'))
    additional_arguments = request.get('args')
    app_launch_command = request.get('cmd')
    platform_id = request.get('platform')
    issue_labels = request.get('issue_labels')
    gestures = request.get('gestures') or '[]'
    stacktrace = request.get('stacktrace')

    crash_data = None
    if job.is_external():
      if not stacktrace:
        raise helpers.EarlyExitException(
            'Stacktrace required for external jobs.', 400)

      if not crash_revision:
        raise helpers.EarlyExitException('Revision required for external jobs.',
                                         400)

      crash_data = stack_analyzer.get_crash_data(
          stacktrace,
          fuzz_target=target_name,
          symbolize_flag=False,
          already_symbolized=True,
          detect_ooms_and_hangs=True)
    elif stacktrace:
      raise helpers.EarlyExitException(
          'Should not specify stacktrace for non-external jobs.', 400)

    testcase_metadata = request.get('metadata', {})
    if testcase_metadata:
      try:
        testcase_metadata = json.loads(testcase_metadata)
      except Exception as e:
        raise helpers.EarlyExitException('Invalid metadata JSON.', 400) from e
      if not isinstance(testcase_metadata, dict):
        raise helpers.EarlyExitException('Metadata is not a JSON object.', 400)
    if issue_labels:
      testcase_metadata['issue_labels'] = issue_labels

    try:
      gestures = ast.literal_eval(gestures)
    except Exception as e:
      raise helpers.EarlyExitException('Failed to parse gestures.', 400) from e

    archive_state = 0
    bundled = False
    file_path_input = ''

    # Certain modifications such as app launch command, issue updates are only
    # allowed for privileged users.
    privileged_user = access.has_access(need_privileged_access=True)
    if not privileged_user:
      if bug_information or bug_summary_update_flag:
        raise helpers.EarlyExitException(
            'You are not privileged to update existing issues.', 400)

      need_privileged_access = utils.string_is_true(
          data_handler.get_value_from_job_definition(job_type,
                                                     'PRIVILEGED_ACCESS'))
      if need_privileged_access:
        raise helpers.EarlyExitException(
            'You are not privileged to run this job type.', 400)

      if app_launch_command:
        raise helpers.EarlyExitException(
            'You are not privileged to run arbitrary launch commands.', 400)

      if (testcase_metadata and
          not _allow_unprivileged_metadata(testcase_metadata)):
        raise helpers.EarlyExitException(
            'You are not privileged to set testcase metadata.', 400)

      if additional_arguments:
        raise helpers.EarlyExitException(
            'You are not privileged to add command-line arguments.', 400)

      if gestures:
        raise helpers.EarlyExitException(
            'You are not privileged to run arbitrary gestures.', 400)

    # TODO(aarya): Remove once AFL is migrated to engine pipeline.
    if target_name:
      additional_arguments = '%TESTCASE%'

    if crash_revision and crash_revision.isdigit():
      crash_revision = int(crash_revision)
    else:
      crash_revision = 0

    if bug_information == '0':  # Auto-recover from this bad input.
      bug_information = None
    if bug_information and not bug_information.isdigit():
      raise helpers.EarlyExitException('Bug is not a number.', 400)

    if not timeout:
      timeout = 0
    elif not timeout.isdigit() or timeout == '0':
      raise helpers.EarlyExitException(
          'Testcase timeout must be a number greater than 0.', 400)
    else:
      timeout = int(timeout)
      if timeout > 120:
        raise helpers.EarlyExitException(
            'Testcase timeout may not be greater than 120 seconds.', 400)

    if retries:
      if retries.isdigit():
        retries = int(retries)
      else:
        retries = None

      if retries is None or retries > MAX_RETRIES:
        raise helpers.EarlyExitException(
            'Testcase retries must be a number less than %d.' % MAX_RETRIES,
            400)
    else:
      retries = None

    job_queue = tasks.queue_for_job(job_type, is_high_end=high_end_job)

    if uploaded_file is not None:
      filename = ''.join([
          x for x in uploaded_file.filename if x not in ' ;/?:@&=+$,{}|<>()\\'
      ])
      key = str(uploaded_file.key())
      if archive.is_archive(filename):
        archive_state = data_types.ArchiveStatus.FUZZED
      if archive_state:
        if multiple_testcases:
          # Create a job to unpack an archive.
          metadata = data_types.BundledArchiveMetadata()
          metadata.blobstore_key = key
          metadata.timeout = timeout
          metadata.job_queue = job_queue
          metadata.job_type = job_type
          metadata.http_flag = http_flag
          metadata.archive_filename = filename
          metadata.uploader_email = email
          metadata.gestures = gestures
          metadata.crash_revision = crash_revision
          metadata.additional_arguments = additional_arguments
          metadata.bug_information = bug_information
          metadata.platform_id = platform_id
          metadata.app_launch_command = app_launch_command
          metadata.fuzzer_name = fuzzer_name
          metadata.overridden_fuzzer_name = fully_qualified_fuzzer_name
          metadata.fuzzer_binary_name = target_name
          metadata.put()

          tasks.add_task(
              'unpack',
              str(metadata.key.id()),
              job_type,
              queue=tasks.queue_for_job(job_type))

          # Create a testcase metadata object to show the user their upload.
          upload_metadata = data_types.TestcaseUploadMetadata()
          upload_metadata.timestamp = datetime.datetime.utcnow()
          upload_metadata.filename = filename
          upload_metadata.blobstore_key = key
          upload_metadata.original_blobstore_key = key
          upload_metadata.status = 'Pending'
          upload_metadata.bundled = True
          upload_metadata.uploader_email = email
          upload_metadata.retries = retries
          upload_metadata.bug_summary_update_flag = bug_summary_update_flag
          upload_metadata.quiet_flag = quiet_flag
          upload_metadata.additional_metadata_string = json.dumps(
              testcase_metadata)
          upload_metadata.bug_information = bug_information
          upload_metadata.put()

          helpers.log('Uploaded multiple testcases.', helpers.VIEW_OPERATION)
          return self.render_json({'multiple': True})

        file_path_input = guess_input_file(uploaded_file, filename)
        if not file_path_input:
          raise helpers.EarlyExitException(
              ("Unable to detect which file to launch. The main file\'s name "
               'must contain either of %s.' % str(RUN_FILE_PATTERNS)), 400)

    else:
      raise helpers.EarlyExitException('Please select a file to upload.', 400)

    testcase_id = data_handler.create_user_uploaded_testcase(
        key,
        key,
        archive_state,
        filename,
        file_path_input,
        timeout,
        job,
        job_queue,
        http_flag,
        gestures,
        additional_arguments,
        bug_information,
        crash_revision,
        email,
        platform_id,
        app_launch_command,
        fuzzer_name,
        fully_qualified_fuzzer_name,
        target_name,
        bundled,
        retries,
        bug_summary_update_flag,
        quiet_flag,
        additional_metadata=testcase_metadata,
        crash_data=crash_data)

    if not quiet_flag:
      testcase = data_handler.get_testcase_by_id(testcase_id)
      issue = issue_tracker_utils.get_issue_for_testcase(testcase)
      if issue:
        report_url = data_handler.TESTCASE_REPORT_URL.format(
            domain=data_handler.get_domain(), testcase_id=testcase_id)

        comment = ('ClusterFuzz is analyzing your testcase. '
                   'Developers can follow the progress at %s.' % report_url)
        issue.save(new_comment=comment)

    helpers.log('Uploaded testcase %s' % testcase_id, helpers.VIEW_OPERATION)
    return self.render_json({'id': '%s' % testcase_id})


class UploadHandler(UploadHandlerCommon, base_handler.GcsUploadHandler):
  """Handler that uploads the testcase file."""

  # pylint: disable=unused-argument
  def before_render_json(self, values, status):
    """Add upload info when the request fails."""
    values['uploadInfo'] = gcs.prepare_blob_upload()._asdict()

  def get_upload(self):
    return base_handler.GcsUploadHandler.get_upload(self)

  @handler.post(handler.FORM, handler.JSON)
  @handler.require_csrf_token
  def post(self):
    return self.do_post()


class NamedBytesIO(io.BytesIO):
  """Named bytesio."""

  def __init__(self, name, value):
    self.name = name
    io.BytesIO.__init__(self, value)


class UploadHandlerOAuth(base_handler.Handler, UploadHandlerCommon):
  """Handler that uploads the testcase file (OAuth)."""

  # pylint: disable=unused-argument
  def before_render_json(self, values, status):
    """Add upload info when the request fails."""
    values['uploadUrl'] = request.host_url + UPLOAD_URL

  def get_upload(self):
    """Get the upload."""
    uploaded_file = request.files.get('file')
    if not uploaded_file:
      raise helpers.EarlyExitException('File upload not found.', 400)

    bytes_io = NamedBytesIO(uploaded_file.filename, uploaded_file.stream.read())
    key = blobs.write_blob(bytes_io)
    return blobs.get_blob_info(key)

  @handler.post(handler.FORM, handler.JSON)
  @handler.oauth
  def post(self, *args):
    return self.do_post()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for untrusted_runner_integration."""

import filecmp
import os
import shutil
import subprocess
import sys
import tempfile

import six

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.tasks import setup
from clusterfuzz._internal.bot.untrusted_runner import \
    environment as untrusted_env
from clusterfuzz._internal.bot.untrusted_runner import config
from clusterfuzz._internal.bot.untrusted_runner import corpus_manager
from clusterfuzz._internal.bot.untrusted_runner import file_host
from clusterfuzz._internal.bot.untrusted_runner import host
from clusterfuzz._internal.bot.untrusted_runner import remote_process_host
from clusterfuzz._internal.bot.untrusted_runner import symbolize_host
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import blobs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import process_handler
from clusterfuzz._internal.system import shell
from clusterfuzz._internal.tests.test_libs import untrusted_runner_helpers

TEST_FILE_CONTENTS = (b'A' * config.FILE_TRANSFER_CHUNK_SIZE +
                      b'B' * config.FILE_TRANSFER_CHUNK_SIZE +
                      b'C' * (config.FILE_TRANSFER_CHUNK_SIZE // 2))

TEST_BUNDLE_BUCKET = 'clusterfuzz-test-bundle'


def _dirs_equal(dircmp):
  if dircmp.left_only or dircmp.right_only or dircmp.diff_files:
    return False

  return all(
      _dirs_equal(sub_dircmp) for sub_dircmp in six.itervalues(dircmp.subdirs))


class UntrustedRunnerIntegrationTest(
    untrusted_runner_helpers.UntrustedRunnerIntegrationTest):
  """Integration tests for untrusted_runner."""

  def assert_dirs_equal(self, dir1, dir2):
    """Assert that 2 dirs are equal."""
    self.assertTrue(_dirs_equal(filecmp.dircmp(dir1, dir2)))

  def setUp(self):
    """Set up."""
    super(UntrustedRunnerIntegrationTest, self).setUp()
    data_types.Config().put()

    environment_string = ('APP_NAME = app\n'
                          'RELEASE_BUILD_BUCKET_PATH = '
                          'gs://clusterfuzz-test-data/test_builds/'
                          'test-build-([0-9]+).zip\n')
    data_types.Job(name='job', environment_string=environment_string).put()

    environment_string = ('RELEASE_BUILD_BUCKET_PATH = '
                          'gs://clusterfuzz-test-data/test_libfuzzer_builds/'
                          'test-libfuzzer-build-([0-9]+).zip\n'
                          'UNPACK_ALL_FUZZ_TARGETS_AND_FILES = True')
    data_types.Job(
        name='libfuzzer_asan_job', environment_string=environment_string).put()

    data_types.Fuzzer(name='fuzzer', data_bundle_name='bundle').put()

    data_types.DataBundle(
        name='bundle', is_local=True, sync_to_worker=True).put()

  def test_run_process(self):
    """Tests remote run_process."""
    expected_output = subprocess.check_output(['ls', '/']).decode('utf-8')

    return_code, _, output = (
        remote_process_host.run_process('ls', current_working_directory='/'))
    self.assertEqual(return_code, 0)
    # run_process adds extra newlines.
    self.assertEqual(output.strip(), expected_output.strip())

  def test_run_and_wait(self):
    """Tests remote run_and_wait."""
    expected_output = subprocess.check_output(['ls', '/'])

    runner = remote_process_host.RemoteProcessRunner('/bin/ls')
    result = runner.run_and_wait(cwd='/')
    self.assertEqual(result.return_code, 0)
    self.assertEqual(result.output, expected_output)

  def test_run_process_env(self):
    """Test environment passing when running processes."""
    environment.set_value('ASAN_OPTIONS', 'host_value')
    runner = remote_process_host.RemoteProcessRunner('/bin/sh', ['-c'])
    result = runner.run_and_wait(['echo $ASAN_OPTIONS'])
    self.assertEqual(result.output, b'host_value\n')

    result = runner.run_and_wait(
        ['echo $ASAN_OPTIONS $UBSAN_OPTIONS'],
        env={
            'UBSAN_OPTIONS': 'ubsan',
            'NOT_PASSED': 'blah'
        })
    self.assertEqual(result.output, b'ubsan\n')

    _, _, output = process_handler.run_process(
        '/bin/sh -c \'echo $ASAN_OPTIONS $MSAN_OPTIONS\'',
        testcase_run=True,
        env_copy={
            'MSAN_OPTIONS': 'msan',
            'NOT_PASSED': 'blah'
        })
    self.assertEqual(output, 'host_value msan')

  def test_create_directory(self):
    """Tests remote create_directory."""
    path = os.path.join(self.tmp_dir, 'dir1')
    self.assertTrue(file_host.create_directory(path))
    self.assertTrue(os.path.isdir(path))

    path = os.path.join(self.tmp_dir, 'dir2', 'dir2')
    self.assertFalse(file_host.create_directory(path))
    self.assertFalse(os.path.exists(path))

    path = os.path.join(self.tmp_dir, 'dir2', 'dir2')
    self.assertTrue(file_host.create_directory(path, True))
    self.assertTrue(os.path.isdir(path))

  def test_remove_directory(self):
    """Tests remote remove_directory."""
    path = os.path.join(self.tmp_dir, 'dir1')
    os.mkdir(path)

    self.assertTrue(file_host.remove_directory(path))
    self.assertFalse(os.path.exists(path))

    os.mkdir(path)
    file_path = os.path.join(path, 'file')
    with open(file_path, 'w') as f:
      f.write('output')

    self.assertTrue(file_host.remove_directory(path, recreate=True))
    self.assertTrue(os.path.isdir(path))
    self.assertFalse(os.path.exists(file_path))

  def test_list_files(self):
    """Tests remote list_files."""
    fuzz_inputs = file_host.rebase_to_worker_root(os.environ['FUZZ_INPUTS'])

    with open(os.path.join(fuzz_inputs, 'a'), 'w') as f:
      f.write('')

    with open(os.path.join(fuzz_inputs, 'b'), 'w') as f:
      f.write('')

    os.mkdir(os.path.join(fuzz_inputs, 'c'))
    with open(os.path.join(fuzz_inputs, 'c', 'c'), 'w') as f:
      f.write('')

    six.assertCountEqual(self, [
        os.path.join(fuzz_inputs, 'a'),
        os.path.join(fuzz_inputs, 'b'),
        os.path.join(fuzz_inputs, 'c'),
    ], file_host.list_files(fuzz_inputs))

    six.assertCountEqual(self, [
        os.path.join(fuzz_inputs, 'a'),
        os.path.join(fuzz_inputs, 'b'),
        os.path.join(fuzz_inputs, 'c', 'c'),
    ], file_host.list_files(fuzz_inputs, recursive=True))

  def test_copy_file_to_worker(self):
    """Tests remote copy_file_to_worker."""
    src_path = os.path.join(self.tmp_dir, 'src')
    with open(src_path, 'wb') as f:
      f.write(TEST_FILE_CONTENTS)

    dest_path = os.path.join(self.tmp_dir, 'dst')
    self.assertTrue(file_host.copy_file_to_worker(src_path, dest_path))

    with open(dest_path, 'rb') as f:
      self.assertEqual(f.read(), TEST_FILE_CONTENTS)

  def test_copy_file_to_worker_intermediate(self):
    """Tests remote copy_file_to_worker creating intermediate paths."""
    src_path = os.path.join(self.tmp_dir, 'src')
    with open(src_path, 'wb') as f:
      f.write(TEST_FILE_CONTENTS)

    dest_path = os.path.join(self.tmp_dir, 'dir1', 'dir2', 'dst')
    self.assertTrue(file_host.copy_file_to_worker(src_path, dest_path))

    with open(dest_path, 'rb') as f:
      self.assertEqual(f.read(), TEST_FILE_CONTENTS)

  def test_write_data_to_worker(self):
    """Tests remote write_data_to_worker."""
    dest_path = os.path.join(self.tmp_dir, 'dst')
    self.assertTrue(
        file_host.write_data_to_worker(b'write_data_to_worker', dest_path))

    with open(dest_path, 'rb') as f:
      self.assertEqual(f.read(), b'write_data_to_worker')

  def test_copy_file_from_worker(self):
    """Tests remote copy_file_from_worker."""
    src_path = os.path.join(self.tmp_dir, 'src')
    with open(src_path, 'wb') as f:
      f.write(TEST_FILE_CONTENTS)

    dest_path = os.path.join(self.tmp_dir, 'dst')
    self.assertTrue(file_host.copy_file_from_worker(src_path, dest_path))

    with open(dest_path, 'rb') as f:
      self.assertEqual(f.read(), TEST_FILE_CONTENTS)

  def test_copy_file_from_worker_does_not_exist(self):
    """Tests remote copy_file_from_worker (does not exist)."""
    src_path = os.path.join(self.tmp_dir, 'DOES_NOT_EXIST')
    dest_path = os.path.join(self.tmp_dir, 'dst')
    self.assertFalse(file_host.copy_file_from_worker(src_path, dest_path))
    self.assertFalse(os.path.exists(dest_path))

  def test_copy_directory_to_worker(self):
    """Tests remote copy_directory_to_worker."""
    src_dir = os.path.join(self.tmp_dir, 'src_dir')
    nested_src_dir = os.path.join(self.tmp_dir, 'src_dir', 'nested')
    os.mkdir(src_dir)
    os.mkdir(nested_src_dir)

    with open(os.path.join(src_dir, 'file1'), 'w') as f:
      f.write('1')

    with open(os.path.join(nested_src_dir, 'file2'), 'w') as f:
      f.write('2')

    dest_dir = os.path.join(self.tmp_dir, 'dst_dir')
    os.mkdir(dest_dir)
    old_file_path = os.path.join(dest_dir, 'old_file')
    with open(old_file_path, 'w') as f:
      f.write('old')

    self.assertTrue(file_host.copy_directory_to_worker(src_dir, dest_dir))
    self.assertTrue(os.path.exists(old_file_path))

    os.remove(old_file_path)
    self.assert_dirs_equal(src_dir, dest_dir)

  def test_copy_directory_to_worker_replace(self):
    """Tests remote copy_directory_to_worker (replacing old dir)"""
    src_dir = os.path.join(self.tmp_dir, 'src_dir')
    os.mkdir(src_dir)

    with open(os.path.join(src_dir, 'file1'), 'w') as f:
      f.write('1')

    dest_dir = os.path.join(self.tmp_dir, 'dst_dir')
    os.mkdir(dest_dir)
    with open(os.path.join(dest_dir, 'old_file'), 'w') as f:
      f.write('old')

    self.assertTrue(
        file_host.copy_directory_to_worker(src_dir, dest_dir, replace=True))
    self.assert_dirs_equal(src_dir, dest_dir)

  def test_setup_regular_build(self):
    """Test setting up a regular build."""
    self._setup_env(job_type='job')
    build = build_manager.setup_build()
    self.assertIsNotNone(build)

    worker_root_dir = os.environ['WORKER_ROOT_DIR']
    expected_build_dir = os.path.join(
        worker_root_dir, 'bot', 'builds', 'clusterfuzz-test-data_test_builds_'
        '2b6ddd7575e9b06b20306183720c65fff3ce318d', 'revisions')
    expected_app_dir = os.path.join(
        worker_root_dir, 'bot', 'builds', 'clusterfuzz-test-data_test_builds_'
        '2b6ddd7575e9b06b20306183720c65fff3ce318d', 'revisions', 'test_build')

    self.assertEqual(
        os.path.join(expected_app_dir, 'app'), os.environ['APP_PATH'])
    self.assertEqual('12345', os.environ['APP_REVISION'])
    self.assertEqual('', os.environ['APP_PATH_DEBUG'])
    self.assertEqual(expected_build_dir, os.environ['BUILD_DIR'])
    self.assertEqual(expected_app_dir, os.environ['APP_DIR'])
    self.assertEqual('', os.environ['FUZZ_TARGET'])

  def test_setup_regular_build_fuzz_target(self):
    """Test setting up a regular build."""
    environment.set_value('TASK_NAME', 'fuzz')
    environment.set_value('TASK_ARGUMENT', 'libFuzzer')

    launcher_dir = os.path.join('src', 'clusterfuzz', '_internal', 'bot',
                                'fuzzers', 'libFuzzer')
    environment.set_value('FUZZER_DIR',
                          os.path.join(os.environ['ROOT_DIR'], launcher_dir))

    self._setup_env(job_type='libfuzzer_asan_job')
    build = build_manager.setup_build(target_weights={})
    self.assertIsNotNone(build)

    worker_root_dir = os.environ['WORKER_ROOT_DIR']
    expected_build_dir = os.path.join(
        worker_root_dir, 'bot', 'builds',
        'clusterfuzz-test-data_test_libfuzzer_builds_'
        '41a87efdd470c6f00e8babf61548bf6c7de57137', 'revisions')

    self.assertEqual('', os.environ['APP_PATH'])
    self.assertEqual('1337', os.environ['APP_REVISION'])
    self.assertEqual('', os.environ['APP_PATH_DEBUG'])
    self.assertEqual(expected_build_dir, os.environ['BUILD_DIR'])
    self.assertEqual('', os.environ['APP_DIR'])
    self.assertEqual('test_fuzzer', os.environ['FUZZ_TARGET'])

  def test_run_process_testcase(self):
    """Test run_process for testcase runs."""
    return_code, _, output = process_handler.run_process(
        '/bin/sh -c \'echo $UNTRUSTED_WORKER\'', testcase_run=True)
    self.assertEqual(return_code, 0)
    self.assertEqual(output, 'True')

    return_code, _, output = process_handler.run_process(
        '/bin/sh -c \'echo $TRUSTED_HOST\'', testcase_run=True)
    self.assertEqual(return_code, 0)
    self.assertEqual(output, '')

  def test_run_process_non_testcase(self):
    """Test run_process for non-testcase runs."""
    return_code, _, output = process_handler.run_process(
        '/bin/sh -c \'echo $TRUSTED_HOST\'', testcase_run=False)
    self.assertEqual(return_code, 0)
    self.assertEqual(output, 'True')

    return_code, _, output = process_handler.run_process(
        '/bin/sh -c \'echo $UNTRUSTED_WORKER\'', testcase_run=False)
    self.assertEqual(return_code, 0)
    self.assertEqual(output, '')

  def test_clear_testcase_directories(self):
    """Test clearing test directories."""
    fuzz_inputs = os.environ['FUZZ_INPUTS']
    worker_fuzz_inputs = file_host.rebase_to_worker_root(fuzz_inputs)

    fuzz_inputs_disk = os.environ['FUZZ_INPUTS_DISK']
    worker_fuzz_inputs_disk = file_host.rebase_to_worker_root(fuzz_inputs_disk)

    with open(os.path.join(worker_fuzz_inputs, 'file'), 'w') as f:
      f.write('blah')

    with open(os.path.join(worker_fuzz_inputs_disk, 'file'), 'w') as f:
      f.write('blah2')

    shell.clear_testcase_directories()
    self.assertEqual(len(os.listdir(worker_fuzz_inputs)), 0)
    self.assertEqual(len(os.listdir(worker_fuzz_inputs_disk)), 0)

  def test_push_testcases_to_worker(self):
    """Test pushing testcases to the worker."""
    fuzz_inputs = os.environ['FUZZ_INPUTS']
    worker_fuzz_inputs = file_host.rebase_to_worker_root(fuzz_inputs)

    with open(os.path.join(worker_fuzz_inputs, 'will_be_replaced'), 'w') as f:
      f.write('blah')

    with open(os.path.join(fuzz_inputs, 'file'), 'w') as f:
      f.write('file')

    subdir = os.path.join(fuzz_inputs, 'subdir')
    os.mkdir(subdir)

    with open(os.path.join(subdir, 'file2'), 'w') as f:
      f.write('file2')

    self.assertTrue(file_host.push_testcases_to_worker())
    self.assert_dirs_equal(fuzz_inputs, worker_fuzz_inputs)

  def test_pulling_testcases_from_worker(self):
    """Test pulling testcases from the worker."""
    fuzz_inputs = os.environ['FUZZ_INPUTS']
    worker_fuzz_inputs = file_host.rebase_to_worker_root(fuzz_inputs)

    with open(os.path.join(fuzz_inputs, 'will_be_replaced'), 'w') as f:
      f.write('blah')

    with open(os.path.join(worker_fuzz_inputs, 'file'), 'w') as f:
      f.write('file')

    subdir = os.path.join(worker_fuzz_inputs, 'subdir')
    os.mkdir(subdir)

    with open(os.path.join(subdir, 'file2'), 'w') as f:
      f.write('file2')

    self.assertTrue(file_host.pull_testcases_from_worker())
    self.assert_dirs_equal(fuzz_inputs, worker_fuzz_inputs)

  def test_setup_testcase(self):
    """Test setup_testcase."""
    job_type = 'job'
    self._setup_env(job_type=job_type)
    fuzz_inputs = os.environ['FUZZ_INPUTS']

    testcase = data_types.Testcase()
    testcase.job_type = job_type
    testcase.absolute_path = os.path.join(fuzz_inputs, 'testcase.ext')

    with tempfile.NamedTemporaryFile() as f:
      f.write(b'contents')
      f.seek(0)
      testcase.fuzzed_keys = blobs.write_blob(f)

    testcase.put()

    file_list, input_directory, testcase_file_path = (
        setup.setup_testcase(testcase, job_type))

    six.assertCountEqual(self, file_list, [
        testcase.absolute_path,
    ])
    self.assertEqual(input_directory, fuzz_inputs)
    self.assertEqual(testcase_file_path, testcase.absolute_path)

    worker_fuzz_inputs = file_host.rebase_to_worker_root(fuzz_inputs)
    self.assert_dirs_equal(fuzz_inputs, worker_fuzz_inputs)

  def test_get_command_line_for_application(self):
    """Test get_command_line_for_application."""
    self._setup_env(job_type='job')
    self.assertIsNotNone(build_manager.setup_build())

    fuzz_inputs = os.environ['FUZZ_INPUTS']
    file_to_run = os.path.join(fuzz_inputs, 'file_to_run')

    os.environ['APP_ARGS'] = '%TESTCASE% %TESTCASE_FILE_URL%'
    command_line = testcase_manager.get_command_line_for_application(
        file_to_run)

    app_path = os.environ['APP_PATH']
    worker_fuzz_inputs = file_host.rebase_to_worker_root(fuzz_inputs)
    worker_file_to_run = os.path.join(worker_fuzz_inputs, 'file_to_run')

    self.assertEqual(
        command_line,
        '%s %s %s' % (app_path, worker_file_to_run,
                      utils.file_path_to_file_url(worker_file_to_run)))

    launcher_path = os.path.join(os.environ['FUZZERS_DIR'], 'test',
                                 'launcher.py')
    os.environ['LAUNCHER_PATH'] = launcher_path
    worker_launcher_path = file_host.rebase_to_worker_root(launcher_path)
    command_line = testcase_manager.get_command_line_for_application(
        file_to_run)
    self.assertEqual(
        command_line,
        '%s %s %s %s %s' % (sys.executable, worker_launcher_path, app_path,
                            worker_file_to_run, worker_file_to_run))

  def test_corpus_sync(self):
    """Test syncing corpus."""
    environment.set_value('CORPUS_BUCKET', 'clusterfuzz-test-data')
    corpus = corpus_manager.RemoteFuzzTargetCorpus('corpus_test_fuzzer',
                                                   'child_fuzzer')
    worker_root = environment.get_value('WORKER_ROOT_DIR')
    test_corpus_directory = os.path.join(worker_root, 'corpus')
    os.mkdir(test_corpus_directory)

    try:
      self.assertTrue(corpus.rsync_to_disk(test_corpus_directory))
      six.assertCountEqual(self, os.listdir(test_corpus_directory),
                           ['123', '456', 'abc'])
    finally:
      if os.path.exists(test_corpus_directory):
        shutil.rmtree(test_corpus_directory, ignore_errors=True)

  def test_symbolize(self):
    """Test symbolize."""
    self._setup_env(job_type='job')
    self.assertIsNotNone(build_manager.setup_build())

    app_dir = environment.get_value('APP_DIR')
    unsymbolized_stacktrace = (
        '#0 0x4f1eb4  ({0}/app+0x4f1eb4)\n'
        '#1 0x4f206e  ({0}/app+0x4f206e)\n').format(app_dir)

    expected_symbolized_stacktrace = (
        '    #0 0x4f1eb4 in Vuln(char*, unsigned long) /usr/local/google/home/'
        'ochang/crashy_binary/test.cc:9:15\n'
        '    #1 0x4f206e in main /usr/local/google/home/ochang/crashy_binary/'
        'test.cc:32:3\n')
    symbolized_stacktrace = symbolize_host.symbolize_stacktrace(
        unsymbolized_stacktrace)
    self.assertEqual(expected_symbolized_stacktrace, symbolized_stacktrace)

  def test_update_data_bundle(self):
    """Test update_data_bundle."""
    self.mock.get_data_bundle_bucket_name.return_value = TEST_BUNDLE_BUCKET
    fuzzer = data_types.Fuzzer.query(data_types.Fuzzer.name == 'fuzzer').get()
    bundle = data_types.DataBundle.query(
        data_types.DataBundle.name == 'bundle').get()

    self.assertTrue(setup.update_data_bundle(fuzzer, bundle))

    data_bundle_directory = file_host.rebase_to_worker_root(
        setup.get_data_bundle_directory('fuzzer'))
    self.assertTrue(os.path.exists(os.path.join(data_bundle_directory, 'a')))
    self.assertTrue(os.path.exists(os.path.join(data_bundle_directory, 'b')))

    self.assertTrue(setup.update_data_bundle(fuzzer, bundle))

  def test_get_fuzz_targets(self):
    """Test get_fuzz_targets."""
    worker_root = os.environ['WORKER_ROOT_DIR']
    worker_test_build_dir = os.path.join(
        worker_root, 'src', 'clusterfuzz', '_internal', 'tests', 'core', 'bot',
        'untrusted_runner', 'test_data', 'test_build')
    fuzz_target_paths = file_host.get_fuzz_targets(worker_test_build_dir)
    six.assertCountEqual(self, [
        os.path.join(worker_test_build_dir, 'do_stuff_fuzzer'),
        os.path.join(worker_test_build_dir, 'target'),
    ], fuzz_target_paths)

  def test_large_message(self):
    """Tests that large messages work."""
    expected_output = b'A' * 1024 * 1024 * 5 + b'\n'

    runner = remote_process_host.RemoteProcessRunner('/usr/bin/python3')
    result = runner.run_and_wait(
        additional_args=['-c', 'print("A"*5*1024*1024)'])
    self.assertEqual(result.return_code, 0)
    self.assertEqual(result.output, expected_output)

  def test_terminate_stale_application_instances(self):
    """Test terminating stale application instances."""
    # TODO(ochang): Improve this test once we use Docker.
    process_handler.terminate_stale_application_instances()

  def test_reset_environment(self):
    """Test resetting environment."""
    environment.set_value('ASAN_OPTIONS', 'saved_options')
    untrusted_env.reset_environment()
    environment.set_value('ASAN_OPTIONS', 'replaced')
    untrusted_env.reset_environment()

    runner = remote_process_host.RemoteProcessRunner('/bin/sh', ['-c'])
    result = runner.run_and_wait(['echo $ASAN_OPTIONS'])
    self.assertEqual(result.output, b'saved_options\n')

  # The "zzz" is a hack to make this test run last (tests are run in
  # alphabetical order).
  # TODO(ochang): Find a better way.
  def test_zzz_update(self):
    """Test updating."""
    host.update_worker()
    self.assertEqual(self.__class__.bot_proc.wait(), 0)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the http_server module."""

import os
import unittest

from pyfakefs import fake_filesystem_unittest
import six

from clusterfuzz._internal.bot.webserver import http_server
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class TestRequestHandler(http_server.RequestHandler):
  """Override methods which would be problematic for testing."""

  class TestWFile(object):

    def __init__(self):
      self.contents = ''

    def write(self, data):
      self.contents += data

  def __init__(self, path):  # pylint: disable=super-init-not-called
    self.response_code = 0
    self.wfile = self.TestWFile()
    self.path = path

  def send_response(self, response_code, _=None):  # pylint: disable=arguments-differ
    self.response_code = response_code

  def send_header(self, *_):  # pylint: disable=arguments-differ
    pass

  def end_headers(self):
    pass


class RequestHandlerTest(fake_filesystem_unittest.TestCase):
  """Tests for the RequestHandler class."""

  def setUp(self):
    """Setup for request handler test."""
    test_utils.set_up_pyfakefs(self, allow_root_user=False)
    helpers.patch_environ(self)

    os.environ['FUZZ_DATA'] = '/data'
    os.environ['FUZZERS_DIR'] = '/fuzzers'
    os.environ['INPUT_DIR'] = '/input'
    self.fs.create_file(
        os.path.join('/input', 'valid.txt'), contents='valid file')
    self.fs.create_file(
        os.path.join('/input', 'unreadable.txt'), contents='unreadable file')
    os.chmod(os.path.join('/input', 'unreadable.txt'), 0)

  def test_nonexistent_file(self):
    """Ensure that we respond with 404 for a nonexistent file."""
    handler = TestRequestHandler('/invalid.txt')
    handler.do_GET()

    self.assertEqual(handler.response_code, 404)
    self.assertEqual(handler.wfile.contents, '')

  def test_unreadable_file(self):
    """Ensure that we respond with 403 for a file we can't read."""
    handler = TestRequestHandler('/unreadable.txt')
    handler.do_GET()

    self.assertEqual(handler.response_code, 403)
    self.assertEqual(handler.wfile.contents, '')

  def test_valid_file(self):
    """Ensure that we respond with 200 and the file contents for a valid one."""
    handler = TestRequestHandler('/valid.txt')
    handler.do_GET()

    self.assertEqual(handler.response_code, 200)
    self.assertEqual(handler.wfile.contents, 'valid file')


class GuessMimeTypeTest(unittest.TestCase):
  """Tests for guess_mime_type."""

  def test_common_mime_types(self):
    """Ensure that we are able to guess common mime types."""
    expected_types_map = {
        'file.html': 'text/html',
        'file.css': 'text/css',
        'file.jpg': 'image/jpeg',
    }
    for filename, expected_value in six.iteritems(expected_types_map):
      self.assertEqual(http_server.guess_mime_type(filename), expected_value)

  def test_invalid_type(self):
    """Ensure that guess_mime_type returns none"""
    self.assertIsNone(
        http_server.guess_mime_type('file.uncommon_extension_79d8f8f6'))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for leaks functions."""

import os
import shutil
import tempfile
import unittest

import mock

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.fuzzing import leak_blacklist
from clusterfuzz._internal.tests.test_libs import test_utils


@test_utils.with_cloud_emulators('datastore')
class LeaksTest(unittest.TestCase):
  """Base class for leaks test cases."""

  def setUp(self):
    self.data_directory = os.path.join(
        os.path.dirname(__file__), 'leak_blacklist_data')
    self.temp_directory = tempfile.mkdtemp()

  def tearDown(self):
    shutil.rmtree(self.temp_directory)

  def _read_test_data(self, filename):
    """Helper function to read files."""
    with open(os.path.join(self.data_directory, filename)) as file_handle:
      return file_handle.read()

  def _add_dummy_leak_testcase(self):
    """Helper function to add a dummy testcase to Testcase database."""
    testcase_item = data_types.Testcase(
        crash_type='Direct-leak', crash_state='test_foo\ntest_bar\n')
    testcase_item.put()
    return testcase_item

  def test_single_leak(self):
    """Test highlighting for report with single direct leak."""
    data = self._read_test_data('single_direct_leak.txt')
    actual_data = leak_blacklist.highlight_first_direct_leak(data)
    expected_data = data
    self.assertEqual(expected_data, actual_data)

  def test_indirect_before_direct_leak(self):
    """Test highlighting when indirect leak precedes first direct leak."""
    data = self._read_test_data('indirect_before_direct_leak.txt')
    actual_data = leak_blacklist.highlight_first_direct_leak(data)
    expected_data = self._read_test_data(
        'indirect_before_direct_leak_highlighted.txt')
    self.assertEqual(expected_data, actual_data)

  def test_multi_direct_leaks(self):
    """Test highlighting for report with multiple direct leaks."""
    data = self._read_test_data('multi_direct_leak.txt')
    actual_data = leak_blacklist.highlight_first_direct_leak(data)
    expected_data = self._read_test_data('multi_direct_leak_expected.txt')
    self.assertEqual(expected_data, actual_data)

  def test_add_to_global_blacklist(self):
    """Test adding element to global blacklist."""
    testcase = self._add_dummy_leak_testcase()
    blacklist_item = leak_blacklist.add_crash_to_global_blacklist_if_needed(
        testcase)
    self.assertTrue(blacklist_item.key.get())

  @mock.patch(
      'clusterfuzz._internal.fuzzing.leak_blacklist.get_local_blacklist_file_path'
  )
  def test_copy_global_to_local_blacklist(self,
                                          mock_get_local_blacklist_file_path):
    """Test copying of global to local blacklist."""
    local_blacklist_file_path = os.path.join(self.temp_directory,
                                             'lsan_suppressions.txt')
    mock_get_local_blacklist_file_path.return_value = local_blacklist_file_path

    testcase = self._add_dummy_leak_testcase()
    blacklist_item = leak_blacklist.add_crash_to_global_blacklist_if_needed(
        testcase)
    self.assertTrue(blacklist_item.key.get())

    # Test that a reproducible leak gets copied to local blacklist file.
    leak_blacklist.copy_global_to_local_blacklist()
    blacklist_function = leak_blacklist.get_leak_function_for_blacklist(
        testcase)
    expected_lsan_suppression_line = (
        leak_blacklist.LSAN_SUPPRESSION_LINE.format(function=blacklist_function)
    )
    self.assertTrue(os.path.isfile(local_blacklist_file_path))
    self.assertIn(expected_lsan_suppression_line,
                  self._read_test_data(local_blacklist_file_path))

    # Test that an excluded reproducible leak is not copied to blacklist file.
    leak_blacklist.copy_global_to_local_blacklist(excluded_testcase=testcase)
    self.assertTrue(os.path.isfile(local_blacklist_file_path))
    self.assertNotIn(expected_lsan_suppression_line,
                     self._read_test_data(local_blacklist_file_path))

  def test_clean_up_global_blacklist(self):
    """Test cleaning of global blacklist."""
    # Test that a reproducible leak is not cleared from blacklist cleanup.
    testcase = self._add_dummy_leak_testcase()
    blacklist_item = leak_blacklist.add_crash_to_global_blacklist_if_needed(
        testcase)
    leak_blacklist.cleanup_global_blacklist()
    self.assertTrue(blacklist_item.key.get())

    # Test that an unreproducible leak is cleared from blacklist cleanup.
    testcase.one_time_crasher_flag = True
    testcase.put()
    leak_blacklist.cleanup_global_blacklist()
    self.assertFalse(blacklist_item.key.get())

    # Flip reproducibility flag and verify that testcase is in blacklist.
    testcase.one_time_crasher_flag = False
    testcase.put()
    blacklist_item = leak_blacklist.add_crash_to_global_blacklist_if_needed(
        testcase)
    self.assertTrue(blacklist_item.key.get())

    # Delete testcase and make sure it is removed from blacklist.
    testcase.key.delete()
    leak_blacklist.cleanup_global_blacklist()
    self.assertFalse(blacklist_item.key.get())


if __name__ == '__main__':
  unittest.main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Simple minimizers for common tasks."""

from . import minimizer


class SinglePassMinimizer(minimizer.Minimizer):  # pylint:disable=abstract-method
  """Do a single pass over the token list."""

  def _execute(self, data):
    """Attempt to remove each token starting from the last one."""
    testcase = minimizer.Testcase(data, self)
    for i in reversed(list(range(len(testcase.tokens)))):
      hypothesis = [i]
      testcase.prepare_test(hypothesis)

    testcase.process()
    return testcase


class EmptyTokenRemover(minimizer.Minimizer):  # pylint:disable=abstract-method
  """Attempt to remove empty tokens."""

  def __init__(self, *args, **kwargs):
    self.is_empty = self._handle_constructor_argument(
        'is_empty', kwargs, default=lambda s: not s.strip())
    minimizer.Minimizer.__init__(self, *args, **kwargs)

  def _execute(self, data):
    """Try to remove all blank tokens, then individual ones."""
    testcase = minimizer.Testcase(data, self)
    tokens = testcase.tokens
    empty_tokens = []

    for i, token in enumerate(tokens):
      if self.is_empty(token):
        empty_tokens.append(i)

    # Try to remove all of them.
    testcase.prepare_test(empty_tokens)

    # The minimizer will automatically skip the following tests if all tokens
    # were removed at once, with the exception of runs started parallel to it.
    for token in empty_tokens:
      testcase.prepare_test([token])

    testcase.process()
    return testcase



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for blame task."""
# pylint: disable=protected-access

import json
import unittest

from clusterfuzz._internal.bot.tasks import blame_task
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.core.bot.tasks.component_revision_patching_test import \
    ComponentRevisionPatchingTest
from clusterfuzz._internal.tests.test_libs import test_utils


class ExtractUrlAndShaFromDepsEntryTest(unittest.TestCase):
  """Test _extract_url_and_sha_from_deps_entry."""

  def test_entry_with_git(self):
    """Ensure that we correctly break deps entries into URLs and SHAs."""
    entry = {
        'url': 'https://site/project.git',
        'rev': '46c07eef042cf06a07a333b41eb50c7d19686b5c'
    }
    url, sha = blame_task._extract_url_and_sha_from_deps_entry(entry)
    self.assertEqual(url, 'https://site/project')
    self.assertEqual(sha, '46c07eef042cf06a07a333b41eb50c7d19686b5c')

  def test_entry_without_git(self):
    entry = {
        'url': 'https://site/project',
        'rev': '0c3c7b36321ce9ddd56c2756ebc4946b2bb98823'
    }
    url, sha = blame_task._extract_url_and_sha_from_deps_entry(entry)
    self.assertEqual(url, 'https://site/project')
    self.assertEqual(sha, '0c3c7b36321ce9ddd56c2756ebc4946b2bb98823')

  def test_malformed_entry(self):
    with self.assertRaises(AssertionError):
      blame_task._extract_url_and_sha_from_deps_entry('malformed')


class ComputeRollsTest(unittest.TestCase):
  """Test _compute_rolls"""

  def test(self):
    """Test."""
    start_revisions_dict = {
        'src/a': {
            'url': 'https://domain/a',
            'rev': '2439bd08ff93d4dce761dd6b825917938bd35a4f'
        },
        'src/b': {
            'url': 'https://domain/b',
            'rev': '2439bd08ff93d4dce761dd6b825917938bd35a4f'
        },
        'src/c/d': {
            'url': 'https://domain/e',
            'rev': '1edde9d2fe203229c895b648fdec355917200ad6'
        }
    }
    end_revisions_dict = {
        'src/a': {
            'url': 'https://domain/a',
            'rev': '2439bd08ff93d4dce761dd6b825917938bd35a4f'
        },
        'src/b': {
            'url': 'https://domain/b',
            'rev': 'e9d71f5ee7c92d6dc9e92ffdad17b8bd49418f98'
        },
        'src/c/d': {
            'url': 'https://domain/f',
            'rev': '5bf1fd927dfb8679496a2e6cf00cbe50c1c87145'
        },
        'src/new': {
            'url': 'https://domain/new',
            'rev': '984844635245b39e82f7f931568393969acd5d31'
        }
    }

    expected_rolls = [
        {
            'dep_path': 'src/b',
            'repo_url': 'https://domain/b',
            'old_revision': '2439bd08ff93d4dce761dd6b825917938bd35a4f',
            'new_revision': 'e9d71f5ee7c92d6dc9e92ffdad17b8bd49418f98',
        },
        {
            'dep_path': 'src/c/d',
            'repo_url': 'https://domain/f',
            'old_revision': '1edde9d2fe203229c895b648fdec355917200ad6',
            'new_revision': '5bf1fd927dfb8679496a2e6cf00cbe50c1c87145',
        },
        {
            'dep_path': 'src/new',
            'repo_url': 'https://domain/new',
            'new_revision': '984844635245b39e82f7f931568393969acd5d31',
        },
    ]

    # Order is not important, but we must sort for comparison to the reference.
    actual_rolls = blame_task._compute_rolls(start_revisions_dict,
                                             end_revisions_dict)
    actual_rolls = sorted(actual_rolls, key=lambda roll: roll['dep_path'])

    self.assertEqual(actual_rolls, expected_rolls)


class FormatComponentRevisionsForPredator(unittest.TestCase):
  """Test _format_component_revisions_for_predator"""

  def test(self):
    """Test."""
    revisions_dict = {
        'src/a': {
            'url': 'https://domain/a',
            'rev': '2439bd08ff93d4dce761dd6b825917938bd35a4f'
        },
        'src/b': {
            'url': 'https://blah/b.git',
            'rev': '2439bd08ff93d4dce761dd6b825917938bd35a4f'
        },
        'src/c/d': {
            'url': 'https://domain/e',
            'rev': '1edde9d2fe203229c895b648fdec355917200ad6'
        }
    }

    expected_result = [
        {
            'dep_path': 'src/a',
            'repo_url': 'https://domain/a',
            'revision': '2439bd08ff93d4dce761dd6b825917938bd35a4f',
        },
        {
            'dep_path': 'src/b',
            'repo_url': 'https://blah/b',
            'revision': '2439bd08ff93d4dce761dd6b825917938bd35a4f',
        },
        {
            'dep_path': 'src/c/d',
            'repo_url': 'https://domain/e',
            'revision': '1edde9d2fe203229c895b648fdec355917200ad6',
        },
    ]

    # Order is not important, but we must sort for comparison to the reference.
    actual_result = blame_task._format_component_revisions_for_predator(
        revisions_dict)
    actual_result = sorted(actual_result, key=lambda roll: roll['dep_path'])

    self.assertEqual(actual_result, expected_result)


class PreparePredatorRequestBodyTest(ComponentRevisionPatchingTest):
  """Test prepare_predator_message."""

  def test_custom_binary(self):
    """Test custom binary."""
    environment.set_value('CUSTOM_BINARY', True)

    testcase = test_utils.create_generic_testcase()
    expected_result = {
        'result': {
            'error_message': 'Not applicable to custom binaries.',
            'feedback_url': '',
            'found': False,
            'project': '',
            'suspected_components': '',
            'suspected_cls': '',
        }
    }

    self.assertIsNone(blame_task._prepare_predator_message(testcase))

    testcase = data_handler.get_testcase_by_id(testcase.key.id())
    actual_result = testcase.get_metadata('predator_result')
    self.assertEqual(actual_result, expected_result)

    environment.remove_key('CUSTOM_BINARY')

  def test_reproducible_no_regression_range(self):
    """Test reproducible with no regression range."""
    testcase = test_utils.create_generic_testcase()
    testcase.one_time_crasher_flag = False
    testcase.regression = ''
    testcase.put()

    expected_result = {
        'result': {
            'error_message':
                'No regression range, wait for regression task to finish.',
            'feedback_url':
                '',
            'found':
                False,
            'project':
                '',
            'suspected_components':
                '',
            'suspected_cls':
                '',
        }
    }

    self.assertIsNone(blame_task._prepare_predator_message(testcase))

    testcase = data_handler.get_testcase_by_id(testcase.key.id())
    actual_result = testcase.get_metadata('predator_result')
    self.assertEqual(actual_result, expected_result)

  def test_reproducible_invalid_regression_range(self):
    """Test reproducible with invalid regression range."""
    testcase = test_utils.create_generic_testcase()
    testcase.one_time_crasher_flag = False
    testcase.regression = 'BAD'
    testcase.put()

    expected_result = {
        'result': {
            'error_message': 'Invalid regression range BAD.',
            'feedback_url': '',
            'found': False,
            'project': '',
            'suspected_components': '',
            'suspected_cls': '',
        }
    }

    self.assertIsNone(blame_task._prepare_predator_message(testcase))

    testcase = data_handler.get_testcase_by_id(testcase.key.id())
    actual_result = testcase.get_metadata('predator_result')
    self.assertEqual(actual_result, expected_result)

  def test_reproducible_regression_range(self):
    """Test reproducible with regression range."""
    testcase = test_utils.create_generic_testcase()
    testcase.crash_revision = 398287
    testcase.regression = '398287:399171'
    testcase.put()

    expected_message = {
        'stack_trace': 'crashy_function()',
        'customized_data': {
            'crash_address':
                '0xdeadbeef',
            'job_type':
                'test_content_shell_drt',
            'dependencies': [{
                'dep_path': 'src/third_party/bidichecker',
                'repo_url': 'https://chromium.googlesource.com/'
                            'external/bidichecker/lib',
                'revision': '97f2aa645b74c28c57eca56992235c79850fa9e0'
            }, {
                'dep_path': 'src/third_party/pdfium',
                'repo_url': 'https://pdfium.googlesource.com/pdfium',
                'revision': 'f7e108b2d0c2f67a143e99693df084bfff7037ec'
            }, {
                'dep_path': 'src/third_party/skia',
                'repo_url': 'https://skia.googlesource.com/skia',
                'revision': 'ee295645bd91fcbe1714847c5fe5341759037cc5'
            }, {
                'dep_path': 'src/v8',
                'repo_url': 'https://chromium.googlesource.com/v8/v8',
                'revision': 'cba1fdd4d72e7c5b874f9eeb07901792f26c871a'
            }, {
                'dep_path': 'src',
                'repo_url': 'https://chromium.googlesource.com/chromium/src',
                'revision': 'e4eb2a57c8667ab31903237e3c316fcaf4afe718'
            }],
            'dependency_rolls': [{
                'dep_path':
                    'src/third_party/gperf',
                'repo_url':
                    'https://chromium.googlesource.com/chromium/deps/gperf',
                'new_revision':
                    '97f2aa645b74c28c57eca56992235c79850fa9e0'
            }, {
                'dep_path': 'src/third_party/pdfium',
                'repo_url': 'https://pdfium.googlesource.com/pdfium',
                'new_revision': '855665d4889853f8ac71519de8ff004dba8eb056',
                'old_revision': 'f7e108b2d0c2f67a143e99693df084bfff7037ec'
            }, {
                'dep_path': 'src/third_party/skia',
                'repo_url': 'https://skia.googlesource.com/skia',
                'new_revision': '4772bd537d153922cd772020e4ad4820090be51a',
                'old_revision': 'ee295645bd91fcbe1714847c5fe5341759037cc5'
            }, {
                'dep_path': 'src/v8',
                'repo_url': 'https://chromium.googlesource.com/v8/v8',
                'new_revision': '3a590058de9b3640f73741b1e95f815f5c089988',
                'old_revision': 'cba1fdd4d72e7c5b874f9eeb07901792f26c871a'
            }, {
                'dep_path': 'src',
                'repo_url': 'https://chromium.googlesource.com/chromium/src',
                'new_revision': '52523d4e58d99cdb768791bf9ac532c917522460',
                'old_revision': 'e4eb2a57c8667ab31903237e3c316fcaf4afe718'
            }],
            'regression_range': {
                'old_revision': 'e4eb2a57c8667ab31903237e3c316fcaf4afe718',
                'new_revision': '52523d4e58d99cdb768791bf9ac532c917522460',
                'dep_path': 'src',
                'repo_url': 'https://chromium.googlesource.com/chromium/src'
            },
            'security_flag':
                False,
            'testcase_id':
                1,
            'sanitizer':
                'ASAN',
            'crash_type':
                'fake type'
        },
        'crash_revision': 'e4eb2a57c8667ab31903237e3c316fcaf4afe718',
        'platform': 'linux',
        'client_id': 'clusterfuzz',
        'signature': 'crashy_function()'
    }

    actual_message = blame_task._prepare_predator_message(testcase)
    actual_message = json.loads(actual_message.data)
    self.assertDictEqual(actual_message, expected_message)

  def test_reproducible_regression_range_with_zero_start_revision(self):
    """Test reproducible with regression range with 0 start."""
    testcase = test_utils.create_generic_testcase()
    testcase.crash_revision = 399171
    testcase.regression = '0:398287'
    testcase.put()

    expected_message = {
        'stack_trace': 'crashy_function()',
        'customized_data': {
            'crash_address':
                '0xdeadbeef',
            'job_type':
                'test_content_shell_drt',
            'dependencies': [{
                'dep_path':
                    'src/third_party/gperf',
                'repo_url':
                    'https://chromium.googlesource.com/chromium/deps/gperf',
                'revision':
                    '97f2aa645b74c28c57eca56992235c79850fa9e0'
            }, {
                'dep_path': 'src/third_party/pdfium',
                'repo_url': 'https://pdfium.googlesource.com/pdfium',
                'revision': '855665d4889853f8ac71519de8ff004dba8eb056'
            }, {
                'dep_path': 'src/third_party/skia',
                'repo_url': 'https://skia.googlesource.com/skia',
                'revision': '4772bd537d153922cd772020e4ad4820090be51a'
            }, {
                'dep_path': 'src/v8',
                'repo_url': 'https://chromium.googlesource.com/v8/v8',
                'revision': '3a590058de9b3640f73741b1e95f815f5c089988'
            }, {
                'dep_path': 'src',
                'repo_url': 'https://chromium.googlesource.com/chromium/src',
                'revision': '52523d4e58d99cdb768791bf9ac532c917522460'
            }],
            'dependency_rolls': [],
            'regression_range': {
                'old_revision': None,
                'new_revision': 'e4eb2a57c8667ab31903237e3c316fcaf4afe718',
                'dep_path': 'src',
                'repo_url': 'https://chromium.googlesource.com/chromium/src'
            },
            'security_flag':
                False,
            'testcase_id':
                1,
            'sanitizer':
                'ASAN',
            'crash_type':
                'fake type'
        },
        'crash_revision': '52523d4e58d99cdb768791bf9ac532c917522460',
        'platform': 'linux',
        'client_id': 'clusterfuzz',
        'signature': 'crashy_function()'
    }

    actual_message = blame_task._prepare_predator_message(testcase)
    actual_message = json.loads(actual_message.data)
    self.assertDictEqual(actual_message, expected_message)

  def test_unreproducible(self):
    """Test unreproducible."""
    testcase = test_utils.create_generic_testcase()
    testcase.one_time_crasher_flag = True
    testcase.crash_revision = 399171
    testcase.regression = 'NA'
    testcase.put()

    expected_message = {
        'stack_trace': 'crashy_function()',
        'customized_data': {
            'crash_address':
                '0xdeadbeef',
            'job_type':
                'test_content_shell_drt',
            'dependencies': [{
                'dep_path':
                    'src/third_party/gperf',
                'repo_url':
                    'https://chromium.googlesource.com/chromium/deps/gperf',
                'revision':
                    '97f2aa645b74c28c57eca56992235c79850fa9e0'
            }, {
                'dep_path': 'src/third_party/pdfium',
                'repo_url': 'https://pdfium.googlesource.com/pdfium',
                'revision': '855665d4889853f8ac71519de8ff004dba8eb056'
            }, {
                'dep_path': 'src/third_party/skia',
                'repo_url': 'https://skia.googlesource.com/skia',
                'revision': '4772bd537d153922cd772020e4ad4820090be51a'
            }, {
                'dep_path': 'src/v8',
                'repo_url': 'https://chromium.googlesource.com/v8/v8',
                'revision': '3a590058de9b3640f73741b1e95f815f5c089988'
            }, {
                'dep_path': 'src',
                'repo_url': 'https://chromium.googlesource.com/chromium/src',
                'revision': '52523d4e58d99cdb768791bf9ac532c917522460'
            }],
            'dependency_rolls': [],
            'regression_range': {
                'old_revision': None,
                'new_revision': None,
                'dep_path': 'src',
                'repo_url': 'https://chromium.googlesource.com/chromium/src'
            },
            'security_flag':
                False,
            'testcase_id':
                1,
            'sanitizer':
                'ASAN',
            'crash_type':
                'fake type'
        },
        'crash_revision': '52523d4e58d99cdb768791bf9ac532c917522460',
        'platform': 'linux',
        'client_id': 'clusterfuzz',
        'signature': 'crashy_function()'
    }

    actual_message = blame_task._prepare_predator_message(testcase)
    actual_message = json.loads(actual_message.data)
    self.assertDictEqual(actual_message, expected_message)


class FilterStacktraceTest(unittest.TestCase):
  """Tests for the _filter_stacktrace helper function."""

  def test_filter_uninteresting_lines(self):
    """Ensure that uninteresting lines are filtered."""
    stacktrace = (
        '[19622:19622:1006/184548.701693:VERBOSE1:bluez_dbus_manager.cc(172)] '
        'Bluetooth not supported.\n'
        '[19622:19622:1006/184548.717837:WARNING:browser_main_loop.cc(294)] '
        '<unknown>: atk-bridge: get_device_events_reply: unknown signature\n'
        'Interesting line 1\n'
        'Interesting line 2\n'
        'Interesting line 3')
    expected_filtered_stacktrace = ('Interesting line 1\n'
                                    'Interesting line 2\n'
                                    'Interesting line 3')
    actual_filtered_stacktrace = blame_task._filter_stacktrace(stacktrace)
    self.assertEqual(expected_filtered_stacktrace, actual_filtered_stacktrace)

  def test_filter_lengthy_stacktrace(self):
    """Ensure that we use the last few lines from a lengthy stacktrace."""
    stacktrace = 'A\n' * 500000
    expected_filtered_stacktrace = 'A\n' * 449999 + 'A'
    actual_filtered_stacktrace = blame_task._filter_stacktrace(stacktrace)
    self.assertEqual(expected_filtered_stacktrace, actual_filtered_stacktrace)
